{
  "cells": [
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "---\n",
        "title: \"LSE DS202W 2025: Week 03 - Student Notebook\"\n",
        "subtitle: \"Linear regression as a machine learning algorithm\"\n",
        "author: \"The DS202 Team\"\n",
        "date: 03 February 2025\n",
        "---\n",
        "\n",
        "## Loading libraries and functions"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 24,
      "metadata": {},
      "outputs": [
        {
          "data": {
            "text/html": [
              "\n",
              "            <div id=\"FovpEz\"></div>\n",
              "            <script type=\"text/javascript\" data-lets-plot-script=\"library\">\n",
              "                if(!window.letsPlotCallQueue) {\n",
              "                    window.letsPlotCallQueue = [];\n",
              "                }; \n",
              "                window.letsPlotCall = function(f) {\n",
              "                    window.letsPlotCallQueue.push(f);\n",
              "                };\n",
              "                (function() {\n",
              "                    var script = document.createElement(\"script\");\n",
              "                    script.type = \"text/javascript\";\n",
              "                    script.src = \"https://cdn.jsdelivr.net/gh/JetBrains/lets-plot@v4.5.2/js-package/distr/lets-plot.min.js\";\n",
              "                    script.onload = function() {\n",
              "                        window.letsPlotCall = function(f) {f();};\n",
              "                        window.letsPlotCallQueue.forEach(function(f) {f();});\n",
              "                        window.letsPlotCallQueue = [];\n",
              "                        \n",
              "                    };\n",
              "                    script.onerror = function(event) {\n",
              "                        window.letsPlotCall = function(f) {};    // noop\n",
              "                        window.letsPlotCallQueue = [];\n",
              "                        var div = document.createElement(\"div\");\n",
              "                        div.style.color = 'darkred';\n",
              "                        div.textContent = 'Error loading Lets-Plot JS';\n",
              "                        document.getElementById(\"FovpEz\").appendChild(div);\n",
              "                    };\n",
              "                    var e = document.getElementById(\"FovpEz\");\n",
              "                    e.appendChild(script);\n",
              "                })()\n",
              "            </script>\n",
              "            "
            ]
          },
          "metadata": {},
          "output_type": "display_data"
        }
      ],
      "source": [
        "import numpy as np\n",
        "import pandas as pd\n",
        "from sklearn.model_selection import train_test_split\n",
        "from sklearn.linear_model import LinearRegression, Lasso\n",
        "from sklearn.metrics import r2_score, root_mean_squared_error\n",
        "import statsmodels.api as sm\n",
        "from lets_plot import *\n",
        "LetsPlot.setup_html()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Before we do anything more\n",
        "\n",
        "Please create a data folder called `data` to store all the different data sets in this course.\n",
        "\n",
        "## The World Values Survey"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 25,
      "metadata": {},
      "outputs": [
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "(86867, 17)\n"
          ]
        }
      ],
      "source": [
        "wvs = pd.read_csv(\"data/wvs-wave-7.csv\")\n",
        "wvs = wvs.dropna()\n",
        "print(wvs.shape)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "We start our machine learning journey with Wave 7 of the World Values Survey (`wvs`), which contains information on 86,867 individuals from 63 countries. We have cleaned the data to only include non-missing values. The columns include:\n",
        "\n",
        "-   `iso3c` 3-letter country iso code\n",
        "-   `satisfaction` 1 to 10 rating of life satisfaction (**the outcome**)\n",
        "-   `social_trust` TRUE / FALSE as to whether or not someone expresses social trust\n",
        "-   `male` respondent is male (reference: female)\n",
        "-   `age` age of respondent\n",
        "-   `post_second_edu` respondent has post-secondary education\n",
        "-   `rural` respondent lives in a rural area (reference: urban)\n",
        "-   `employment` categorical employment variable (try `count(wvs, employment)`)\n",
        "-   `financial_situ` 1 to 10 rating of respondent's financial situation\n",
        "-   `married` the respondent is married (reference: other marital status)\n",
        "-   `relig_import` 1 to 10 rating of how much importance the respondent attaches to religion.\n",
        "-   `better_living` trichotomous rating of whether or not the respondent sees their lives as better off than their parents (try `count(wvs, better_living)`)\n",
        "-   `no_food`: In the last 12 months, how often have your or your family gone without enough food to eat? (Sometimes / Often = TRUE, Rarely / Never = FALSE)\n",
        "-   `no_safety` In the last 12 months, how often have your or your family felt unsafe from crime in your home? (Sometimes / Often = TRUE, Rarely / Never = FALSE)\n",
        "-   `no_medical` In the last 12 months, how often have your or your family gone without medicine or medical treatment that you needed? (Sometimes / Often = TRUE, Rarely / Never = FALSE)\n",
        "-   `no_cash` In the last 12 months, how often have your or your family gone without a cash income? (Sometimes / Often = TRUE, Rarely / Never = FALSE)\n",
        "-   `no_shelter` In the last 12 months, how often have your or your family gone without a safe shelter over your head? (Sometimes / Often = TRUE, Rarely / Never = FALSE)\n",
        "\n",
        "👉 **NOTE:** With variables such as `no_food`, we recoded the original question mainly for the sake of simplicity. This simplification can be useful as it reduces the number of parameters in our model. However, there may be distinct differences between each level that may produce different results.\n",
        "\n",
        "If you find that your laptop is unable to handle the full data set without running slowly, try experimenting with the following code. This takes a random sample of the data, stratifying by country so the sampling algorithm doesn't take more data from one country and less from another."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 6,
      "metadata": {},
      "outputs": [],
      "source": [
        "wvs = wvs.groupby(\"iso3c\").sample(frac=0.25,random_state=123)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Understanding life satisfaction: some exploratory data analysis (EDA) (5 minutes)\n",
        "\n",
        "Here are a couple of graphs that tell a few stories about `wvs`.\n",
        "\n",
        "### Median life satisfaction appears to be relatively high in the sample (7 out of 10)"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 7,
      "metadata": {},
      "outputs": [
        {
          "data": {
            "text/html": [
              "   <div id=\"2JYMCH\"></div>\n",
              "   <script type=\"text/javascript\" data-lets-plot-script=\"plot\">\n",
              "   \n",
              "   (function() {\n",
              "   // ----------\n",
              "   \n",
              "   var containerDiv = document.getElementById(\"2JYMCH\");\n",
              "   var observer = new ResizeObserver(function(entries) {\n",
              "       for (let entry of entries) {\n",
              "           var width = containerDiv.clientWidth\n",
              "           if (entry.contentBoxSize && width > 0) {\n",
              "           \n",
              "               // Render plot\n",
              "               if (observer) {\n",
              "                   observer.disconnect();\n",
              "                   observer = null;\n",
              "               }\n",
              "\n",
              "               var plotSpec={\n",
              "\"data\":{\n",
              "},\n",
              "\"mapping\":{\n",
              "\"x\":\"satisfaction\"\n",
              "},\n",
              "\"data_meta\":{\n",
              "\"series_annotations\":[{\n",
              "\"type\":\"str\",\n",
              "\"column\":\"iso3c\"\n",
              "},{\n",
              "\"type\":\"int\",\n",
              "\"column\":\"satisfaction\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"social_trust\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"male\"\n",
              "},{\n",
              "\"type\":\"int\",\n",
              "\"column\":\"age\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"post_second_edu\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"rural\"\n",
              "},{\n",
              "\"type\":\"str\",\n",
              "\"column\":\"employment\"\n",
              "},{\n",
              "\"type\":\"int\",\n",
              "\"column\":\"financial_situ\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"married\"\n",
              "},{\n",
              "\"type\":\"int\",\n",
              "\"column\":\"relig_import\"\n",
              "},{\n",
              "\"type\":\"str\",\n",
              "\"column\":\"better_living\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"no_food\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"no_safety\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"no_medical\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"no_cash\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"no_shelter\"\n",
              "}]\n",
              "},\n",
              "\"theme\":{\n",
              "\"name\":\"minimal\",\n",
              "\"panel_grid_minor\":{\n",
              "\"blank\":true\n",
              "},\n",
              "\"panel_grid_major_x\":{\n",
              "\"blank\":true\n",
              "}\n",
              "},\n",
              "\"caption\":{\n",
              "\"text\":\"Note: dotted line represents median life satisfaction\"\n",
              "},\n",
              "\"guides\":{\n",
              "\"x\":{\n",
              "\"title\":\"Life satisfaction\"\n",
              "},\n",
              "\"y\":{\n",
              "\"title\":\"Number of respondents\"\n",
              "}\n",
              "},\n",
              "\"kind\":\"plot\",\n",
              "\"scales\":[{\n",
              "\"aesthetic\":\"x\",\n",
              "\"breaks\":[1.0,7.0,10.0]\n",
              "}],\n",
              "\"layers\":[{\n",
              "\"geom\":\"histogram\",\n",
              "\"mapping\":{\n",
              "},\n",
              "\"data_meta\":{\n",
              "},\n",
              "\"bins\":10.0,\n",
              "\"fill\":\"grey\",\n",
              "\"colour\":\"black\",\n",
              "\"alpha\":0.5,\n",
              "\"data\":{\n",
              "\"..count..\":[550.0,298.0,657.0,908.0,2490.0,2484.0,3797.0,4472.0,2297.0,3762.0],\n",
              "\"satisfaction\":[0.883,1.909,2.935,3.9610000000000003,4.987,6.013,7.039000000000001,8.065000000000001,9.091000000000001,10.117]\n",
              "}\n",
              "},{\n",
              "\"geom\":\"vline\",\n",
              "\"mapping\":{\n",
              "},\n",
              "\"data_meta\":{\n",
              "},\n",
              "\"xintercept\":7.0,\n",
              "\"linetype\":\"dashed\",\n",
              "\"size\":2.0,\n",
              "\"colour\":\"red\",\n",
              "\"data\":{\n",
              "}\n",
              "}],\n",
              "\"metainfo_list\":[],\n",
              "\"spec_id\":\"1\"\n",
              "};\n",
              "               window.letsPlotCall(function() {\n",
              "       \n",
              "               var toolbar = null;\n",
              "               var plotContainer = containerDiv;               \n",
              "               \n",
              "                   var options = {\n",
              "                       sizing: {\n",
              "                           width_mode: \"min\",\n",
              "                           height_mode: \"scaled\",\n",
              "                           width: width\n",
              "                       }\n",
              "                   };\n",
              "                   var fig = LetsPlot.buildPlotFromProcessedSpecs(plotSpec, -1, -1, plotContainer, options);\n",
              "                   if (toolbar) {\n",
              "                     toolbar.bind(fig);\n",
              "                   }\n",
              "               });\n",
              "               \n",
              "               break;\n",
              "           }\n",
              "       }\n",
              "   });\n",
              "   \n",
              "   observer.observe(containerDiv);\n",
              "   \n",
              "   // ----------\n",
              "   })();\n",
              "   \n",
              "   </script>"
            ],
            "text/plain": [
              "<lets_plot.plot.core.PlotSpec at 0x1cda73d1760>"
            ]
          },
          "execution_count": 7,
          "metadata": {},
          "output_type": "execute_result"
        }
      ],
      "source": [
        "med_satisfaction = np.median(wvs[\"satisfaction\"])\n",
        "\n",
        "(\n",
        "  ggplot(wvs, aes(\"satisfaction\")) +\n",
        "  geom_histogram(fill = \"grey\", colour = \"black\", bins = 10, alpha = 0.5) +\n",
        "  geom_vline(xintercept = med_satisfaction, linetype = \"dashed\", size = 2, colour = \"red\") +\n",
        "  scale_x_continuous(breaks = [1, med_satisfaction, 10]) +\n",
        "  theme_minimal() +\n",
        "  theme(panel_grid_minor = element_blank(),\n",
        "        panel_grid_major_x=element_blank()) +\n",
        "  labs(x = \"Life satisfaction\", y = \"Number of respondents\",\n",
        "       caption = \"Note: dotted line represents median life satisfaction\")\n",
        ")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Life satisfaction tracks positively with financial situation"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 8,
      "metadata": {},
      "outputs": [
        {
          "data": {
            "text/html": [
              "   <div id=\"c37aab\"></div>\n",
              "   <script type=\"text/javascript\" data-lets-plot-script=\"plot\">\n",
              "   \n",
              "   (function() {\n",
              "   // ----------\n",
              "   \n",
              "   var containerDiv = document.getElementById(\"c37aab\");\n",
              "   var observer = new ResizeObserver(function(entries) {\n",
              "       for (let entry of entries) {\n",
              "           var width = containerDiv.clientWidth\n",
              "           if (entry.contentBoxSize && width > 0) {\n",
              "           \n",
              "               // Render plot\n",
              "               if (observer) {\n",
              "                   observer.disconnect();\n",
              "                   observer = null;\n",
              "               }\n",
              "\n",
              "               var plotSpec={\n",
              "\"data\":{\n",
              "\"financial_situ\":[10.0,8.0,7.0,7.0,5.0,6.0,6.0,6.0,8.0,9.0,5.0,5.0,8.0,8.0,5.0,9.0,5.0,7.0,7.0,1.0,4.0,9.0,6.0,4.0,7.0,3.0,4.0,6.0,4.0,10.0,6.0,1.0,8.0,7.0,3.0,3.0,4.0,3.0,10.0,5.0,3.0,1.0,5.0,9.0,2.0,3.0,2.0,8.0,4.0,1.0,4.0,2.0,6.0,2.0,1.0,1.0,1.0,1.0,10.0,10.0,1.0,5.0,2.0,3.0,2.0,2.0,4.0,2.0,7.0,9.0,5.0,3.0,6.0,3.0,2.0,1.0,3.0,10.0,10.0,9.0,7.0,8.0,8.0,2.0,4.0,5.0,4.0,9.0,10.0,7.0,6.0,10.0,10.0,6.0,7.0,9.0,9.0,8.0,8.0,9.0],\n",
              "\"satisfaction\":[10.0,8.0,7.0,8.0,5.0,7.0,6.0,8.0,9.0,9.0,6.0,7.0,7.0,10.0,8.0,8.0,10.0,9.0,6.0,1.0,5.0,10.0,5.0,7.0,10.0,5.0,6.0,10.0,4.0,9.0,9.0,10.0,6.0,5.0,7.0,3.0,8.0,4.0,8.0,9.0,6.0,5.0,4.0,7.0,3.0,8.0,5.0,5.0,3.0,3.0,10.0,4.0,4.0,2.0,6.0,2.0,7.0,4.0,5.0,7.0,8.0,3.0,8.0,10.0,6.0,10.0,9.0,7.0,4.0,6.0,1.0,2.0,3.0,9.0,1.0,9.0,1.0,6.0,1.0,5.0,3.0,4.0,3.0,9.0,1.0,2.0,2.0,4.0,2.0,1.0,2.0,3.0,4.0,1.0,2.0,1.0,3.0,2.0,1.0,2.0],\n",
              "\"count\":[1630.0,1413.0,1159.0,1050.0,947.0,779.0,724.0,605.0,596.0,594.0,571.0,567.0,498.0,495.0,438.0,391.0,379.0,368.0,356.0,337.0,331.0,306.0,297.0,293.0,293.0,260.0,258.0,257.0,246.0,220.0,206.0,192.0,186.0,185.0,179.0,175.0,175.0,162.0,160.0,157.0,156.0,151.0,139.0,131.0,115.0,114.0,113.0,107.0,93.0,93.0,93.0,90.0,89.0,86.0,86.0,81.0,69.0,69.0,67.0,66.0,65.0,63.0,61.0,60.0,60.0,57.0,57.0,56.0,55.0,52.0,48.0,45.0,39.0,38.0,36.0,36.0,36.0,35.0,33.0,32.0,31.0,30.0,28.0,25.0,22.0,19.0,17.0,16.0,15.0,14.0,14.0,12.0,12.0,9.0,8.0,8.0,8.0,7.0,7.0,6.0]\n",
              "},\n",
              "\"mapping\":{\n",
              "\"x\":\"financial_situ\",\n",
              "\"y\":\"satisfaction\",\n",
              "\"size\":\"count\"\n",
              "},\n",
              "\"data_meta\":{\n",
              "\"series_annotations\":[{\n",
              "\"type\":\"int\",\n",
              "\"column\":\"financial_situ\"\n",
              "},{\n",
              "\"type\":\"int\",\n",
              "\"column\":\"satisfaction\"\n",
              "},{\n",
              "\"type\":\"int\",\n",
              "\"column\":\"count\"\n",
              "}]\n",
              "},\n",
              "\"theme\":{\n",
              "\"legend_position\":\"bottom\",\n",
              "\"panel_background\":{\n",
              "\"fill\":\"white\",\n",
              "\"blank\":false\n",
              "},\n",
              "\"panel_grid\":{\n",
              "\"blank\":true\n",
              "}\n",
              "},\n",
              "\"guides\":{\n",
              "\"x\":{\n",
              "\"title\":\"Financial situation\"\n",
              "},\n",
              "\"y\":{\n",
              "\"title\":\"Life satisfaction\"\n",
              "}\n",
              "},\n",
              "\"kind\":\"plot\",\n",
              "\"scales\":[{\n",
              "\"aesthetic\":\"x\",\n",
              "\"breaks\":[2.0,4.0,6.0,8.0,10.0]\n",
              "},{\n",
              "\"aesthetic\":\"y\",\n",
              "\"breaks\":[2.0,4.0,6.0,8.0,10.0]\n",
              "}],\n",
              "\"layers\":[{\n",
              "\"geom\":\"point\",\n",
              "\"mapping\":{\n",
              "},\n",
              "\"data_meta\":{\n",
              "},\n",
              "\"data\":{\n",
              "}\n",
              "}],\n",
              "\"metainfo_list\":[],\n",
              "\"spec_id\":\"2\"\n",
              "};\n",
              "               window.letsPlotCall(function() {\n",
              "       \n",
              "               var toolbar = null;\n",
              "               var plotContainer = containerDiv;               \n",
              "               \n",
              "                   var options = {\n",
              "                       sizing: {\n",
              "                           width_mode: \"min\",\n",
              "                           height_mode: \"scaled\",\n",
              "                           width: width\n",
              "                       }\n",
              "                   };\n",
              "                   var fig = LetsPlot.buildPlotFromProcessedSpecs(plotSpec, -1, -1, plotContainer, options);\n",
              "                   if (toolbar) {\n",
              "                     toolbar.bind(fig);\n",
              "                   }\n",
              "               });\n",
              "               \n",
              "               break;\n",
              "           }\n",
              "       }\n",
              "   });\n",
              "   \n",
              "   observer.observe(containerDiv);\n",
              "   \n",
              "   // ----------\n",
              "   })();\n",
              "   \n",
              "   </script>"
            ],
            "text/plain": [
              "<lets_plot.plot.core.PlotSpec at 0x1cda7400230>"
            ]
          },
          "execution_count": 8,
          "metadata": {},
          "output_type": "execute_result"
        }
      ],
      "source": [
        "wvs_count = wvs.value_counts([\"financial_situ\", \"satisfaction\"]).reset_index()\n",
        "\n",
        "(\n",
        "  ggplot(wvs_count, aes(\"financial_situ\", \"satisfaction\", size = \"count\")) +\n",
        "  geom_point() +\n",
        "  scale_x_continuous(breaks = np.arange(2, 12, 2)) +\n",
        "  scale_y_continuous(breaks = np.arange(2, 12, 2)) +\n",
        "  theme(panel_grid = element_blank(),\n",
        "        panel_background = element_rect(fill = \"white\"),\n",
        "        legend_position = \"bottom\") +\n",
        "  labs(x = \"Financial situation\", y = \"Life satisfaction\")\n",
        ")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### The median individual with post-secondary education has a higher life satisfaction than the median individual without"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 9,
      "metadata": {},
      "outputs": [
        {
          "data": {
            "text/html": [
              "   <div id=\"19kWVV\"></div>\n",
              "   <script type=\"text/javascript\" data-lets-plot-script=\"plot\">\n",
              "   \n",
              "   (function() {\n",
              "   // ----------\n",
              "   \n",
              "   var containerDiv = document.getElementById(\"19kWVV\");\n",
              "   var observer = new ResizeObserver(function(entries) {\n",
              "       for (let entry of entries) {\n",
              "           var width = containerDiv.clientWidth\n",
              "           if (entry.contentBoxSize && width > 0) {\n",
              "           \n",
              "               // Render plot\n",
              "               if (observer) {\n",
              "                   observer.disconnect();\n",
              "                   observer = null;\n",
              "               }\n",
              "\n",
              "               var plotSpec={\n",
              "\"data\":{\n",
              "},\n",
              "\"mapping\":{\n",
              "\"x\":\"post_second_edu\",\n",
              "\"y\":\"satisfaction\",\n",
              "\"fill\":\"post_second_edu\"\n",
              "},\n",
              "\"data_meta\":{\n",
              "\"series_annotations\":[{\n",
              "\"type\":\"str\",\n",
              "\"column\":\"iso3c\"\n",
              "},{\n",
              "\"type\":\"int\",\n",
              "\"column\":\"satisfaction\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"social_trust\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"male\"\n",
              "},{\n",
              "\"type\":\"int\",\n",
              "\"column\":\"age\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"post_second_edu\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"rural\"\n",
              "},{\n",
              "\"type\":\"str\",\n",
              "\"column\":\"employment\"\n",
              "},{\n",
              "\"type\":\"int\",\n",
              "\"column\":\"financial_situ\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"married\"\n",
              "},{\n",
              "\"type\":\"int\",\n",
              "\"column\":\"relig_import\"\n",
              "},{\n",
              "\"type\":\"str\",\n",
              "\"column\":\"better_living\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"no_food\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"no_safety\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"no_medical\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"no_cash\"\n",
              "},{\n",
              "\"type\":\"bool\",\n",
              "\"column\":\"no_shelter\"\n",
              "}]\n",
              "},\n",
              "\"theme\":{\n",
              "\"name\":\"minimal\",\n",
              "\"legend_position\":\"none\",\n",
              "\"panel_grid_minor\":{\n",
              "\"blank\":true\n",
              "},\n",
              "\"panel_grid_major_x\":{\n",
              "\"blank\":true\n",
              "}\n",
              "},\n",
              "\"guides\":{\n",
              "\"x\":{\n",
              "\"title\":\"Post-secondary education?\"\n",
              "},\n",
              "\"y\":{\n",
              "\"title\":\"Life satisfaction\"\n",
              "}\n",
              "},\n",
              "\"kind\":\"plot\",\n",
              "\"scales\":[],\n",
              "\"layers\":[{\n",
              "\"geom\":\"boxplot\",\n",
              "\"mapping\":{\n",
              "},\n",
              "\"data_meta\":{\n",
              "},\n",
              "\"data\":{\n",
              "\"..middle..\":[7.0,8.0],\n",
              "\"..lower..\":[6.0,6.0],\n",
              "\"..upper..\":[9.0,9.0],\n",
              "\"..ymin..\":[2.0,2.0],\n",
              "\"..ymax..\":[10.0,10.0],\n",
              "\"post_second_edu\":[false,true]\n",
              "}\n",
              "},{\n",
              "\"geom\":\"point\",\n",
              "\"stat\":\"boxplot_outlier\",\n",
              "\"mapping\":{\n",
              "},\n",
              "\"show_legend\":false,\n",
              "\"data_meta\":{\n",
              "},\n",
              "\"data\":{\n",
              "\"..middle..\":[7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,7.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0,8.0],\n",
              "\"..lower..\":[6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0,6.0],\n",
              "\"..upper..\":[9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0,9.0],\n",
              "\"..ymin..\":[2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0,2.0],\n",
              "\"..ymax..\":[10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0,10.0],\n",
              "\"post_second_edu\":[false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true,true],\n",
              "\"satisfaction\":[1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0]\n",
              "}\n",
              "}],\n",
              "\"metainfo_list\":[],\n",
              "\"spec_id\":\"3\"\n",
              "};\n",
              "               window.letsPlotCall(function() {\n",
              "       \n",
              "               var toolbar = null;\n",
              "               var plotContainer = containerDiv;               \n",
              "               \n",
              "                   var options = {\n",
              "                       sizing: {\n",
              "                           width_mode: \"min\",\n",
              "                           height_mode: \"scaled\",\n",
              "                           width: width\n",
              "                       }\n",
              "                   };\n",
              "                   var fig = LetsPlot.buildPlotFromProcessedSpecs(plotSpec, -1, -1, plotContainer, options);\n",
              "                   if (toolbar) {\n",
              "                     toolbar.bind(fig);\n",
              "                   }\n",
              "               });\n",
              "               \n",
              "               break;\n",
              "           }\n",
              "       }\n",
              "   });\n",
              "   \n",
              "   observer.observe(containerDiv);\n",
              "   \n",
              "   // ----------\n",
              "   })();\n",
              "   \n",
              "   </script>"
            ],
            "text/plain": [
              "<lets_plot.plot.core.PlotSpec at 0x1cda74011c0>"
            ]
          },
          "execution_count": 9,
          "metadata": {},
          "output_type": "execute_result"
        }
      ],
      "source": [
        "(\n",
        "  ggplot(wvs, aes(\"post_second_edu\", \"satisfaction\", fill = \"post_second_edu\")) +\n",
        "  geom_boxplot() +\n",
        "  theme_minimal() +\n",
        "  theme(panel_grid_minor = element_blank(),\n",
        "        panel_grid_major_x = element_blank(),\n",
        "        legend_position=\"none\") +\n",
        "  labs(x = \"Post-secondary education?\", y = \"Life satisfaction\")\n",
        ")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Understanding life satisfaction: the hypothesis-testing approach (5 minutes)\n",
        "\n",
        "Why do some people have a higher life satisfaction than others? This is one question that a quantitative social scientist might answer by exploring the magnitude and precision of a series of variables. Suppose we hypothesised that individuals with post-secondary level education have greater life satisfaction relative to those who do not. We can estimate a linear regression model by using `satisfaction` as the dependent variable and `post_second_edu` as the independent variable.\n",
        "\n",
        "To estimate a linear regression with interpretable output in Python, we can use the `sm.OLS` function, which requires two things:\n",
        "\n",
        "-   One or more features plus a constant\n",
        "-   An outcome\n",
        "\n",
        "Let's do this now. We can print the summary method to get information on the coefficient estimate for `post_second_edu`."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 10,
      "metadata": {},
      "outputs": [
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "                            OLS Regression Results                            \n",
            "==============================================================================\n",
            "Dep. Variable:           satisfaction   R-squared:                       0.002\n",
            "Model:                            OLS   Adj. R-squared:                  0.002\n",
            "Method:                 Least Squares   F-statistic:                     42.24\n",
            "Date:                Thu, 30 Jan 2025   Prob (F-statistic):           8.23e-11\n",
            "Time:                        10:56:27   Log-Likelihood:                -47959.\n",
            "No. Observations:               21715   AIC:                         9.592e+04\n",
            "Df Residuals:                   21713   BIC:                         9.594e+04\n",
            "Df Model:                           1                                         \n",
            "Covariance Type:            nonrobust                                         \n",
            "===================================================================================\n",
            "                      coef    std err          t      P>|t|      [0.025      0.975]\n",
            "-----------------------------------------------------------------------------------\n",
            "const               7.0588      0.018    387.461      0.000       7.023       7.094\n",
            "post_second_edu     0.2071      0.032      6.500      0.000       0.145       0.270\n",
            "==============================================================================\n",
            "Omnibus:                     1373.715   Durbin-Watson:                   1.793\n",
            "Prob(Omnibus):                  0.000   Jarque-Bera (JB):             1650.277\n",
            "Skew:                          -0.673   Prob(JB):                         0.00\n",
            "Kurtosis:                       3.106   Cond. No.                         2.41\n",
            "==============================================================================\n",
            "\n",
            "Notes:\n",
            "[1] Standard Errors assume that the covariance matrix of the errors is correctly specified.\n"
          ]
        }
      ],
      "source": [
        "# Convert post_second_edu to float\n",
        "X = wvs[\"post_second_edu\"].astype(float)\n",
        "# Add a constant\n",
        "X = sm.add_constant(X)\n",
        "\n",
        "# Isolate the feature column\n",
        "y = wvs[\"satisfaction\"]\n",
        "\n",
        "# Build an OLS and print its output\n",
        "model_univ = sm.OLS(y,X).fit()\n",
        "print(model_univ.summary())"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "We see that individuals with post-secondary education have a positive and statistically significant (p \\< 0.001) increase in life satisfaction of about 0.2 points.\n",
        "\n",
        "👉 **NOTE:** The process of hypothesis testing is obviously more involved when using observational data than is portrayed by this simple example. Control variables will almost always be incorporated and, increasingly, identification strategies will be used to uncover causal effects. The end result, however, will involve as rigorous an attempt at falsifying a hypothesis as can be provided with the data.\n",
        "\n",
        "For an example of how multivariate regression is used, we can run the following code."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 11,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Create a function that normalises variables\n",
        "def normalise(var):\n",
        "  return (var-var.mean())/var.std()\n",
        "\n",
        "# Create a data frame of features\n",
        "X = wvs.drop([\"satisfaction\",\"iso3c\"],axis=1)\n",
        "\n",
        "# Identify which features are floats, categories and booleians\n",
        "X_float = X.filter(items=[\"age\",\"financial_situ\",\"relig_import\"],axis=1)\n",
        "X_cat = X.filter(items=[\"employment\",\"better_living\"],axis=1)\n",
        "X_bool = X[X.columns[~X.columns.isin(X_float+X_cat)]]\n",
        "\n",
        "# Normalise numeric features\n",
        "X_float = X_float.apply(lambda x: normalise(x),axis=0)\n",
        "\n",
        "# Get dummies from categorical features\n",
        "X_cat = pd.get_dummies(X_cat,drop_first=True, dtype=int)\n",
        "\n",
        "# Convert booleian features to integers\n",
        "X_bool = X_bool.astype(int)\n",
        "\n",
        "# Concatenate to final data frame\n",
        "X = pd.concat([X_float,X_cat,X_bool],axis=1)\n",
        "\n",
        "# Add a constant\n",
        "X = sm.add_constant(X)\n",
        "\n",
        "# ISOLATING OUR OUTCOME \n",
        "y = wvs[\"satisfaction\"]\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 12,
      "metadata": {},
      "outputs": [
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "                            OLS Regression Results                            \n",
            "==============================================================================\n",
            "Dep. Variable:           satisfaction   R-squared:                       0.324\n",
            "Model:                            OLS   Adj. R-squared:                  0.323\n",
            "Method:                 Least Squares   F-statistic:                     472.3\n",
            "Date:                Thu, 30 Jan 2025   Prob (F-statistic):               0.00\n",
            "Time:                        10:56:34   Log-Likelihood:                -43730.\n",
            "No. Observations:               21715   AIC:                         8.751e+04\n",
            "Df Residuals:                   21692   BIC:                         8.769e+04\n",
            "Df Model:                          22                                         \n",
            "Covariance Type:            nonrobust                                         \n",
            "=================================================================================================\n",
            "                                    coef    std err          t      P>|t|      [0.025      0.975]\n",
            "-------------------------------------------------------------------------------------------------\n",
            "const                             7.1578      0.041    176.680      0.000       7.078       7.237\n",
            "age                               0.0019      0.016      0.118      0.906      -0.030       0.034\n",
            "financial_situ                    1.1439      0.013     85.712      0.000       1.118       1.170\n",
            "relig_import                      0.2020      0.013     15.168      0.000       0.176       0.228\n",
            "employment_House wife/husband    -0.0497      0.044     -1.129      0.259      -0.136       0.037\n",
            "employment_Other                  0.0397      0.114      0.348      0.728      -0.184       0.263\n",
            "employment_Part time              0.1444      0.047      3.080      0.002       0.052       0.236\n",
            "employment_Retired               -0.0248      0.048     -0.514      0.608      -0.119       0.070\n",
            "employment_Self employed          0.0224      0.040      0.563      0.573      -0.055       0.100\n",
            "employment_Student               -0.0373      0.061     -0.608      0.543      -0.158       0.083\n",
            "employment_Unemployed            -0.1534      0.050     -3.043      0.002      -0.252      -0.055\n",
            "better_living_Better off          0.0211      0.029      0.727      0.467      -0.036       0.078\n",
            "better_living_Worse off          -0.4207      0.039    -10.845      0.000      -0.497      -0.345\n",
            "social_trust                      0.0517      0.030      1.720      0.085      -0.007       0.111\n",
            "male                             -0.0582      0.027     -2.190      0.029      -0.110      -0.006\n",
            "post_second_edu                  -0.0115      0.028     -0.411      0.681      -0.066       0.043\n",
            "rural                            -0.0191      0.028     -0.692      0.489      -0.073       0.035\n",
            "married                           0.1937      0.028      6.980      0.000       0.139       0.248\n",
            "no_food                          -0.2239      0.041     -5.414      0.000      -0.305      -0.143\n",
            "no_safety                        -0.0114      0.036     -0.313      0.754      -0.083       0.060\n",
            "no_medical                       -0.1025      0.037     -2.802      0.005      -0.174      -0.031\n",
            "no_cash                           0.0399      0.034      1.171      0.242      -0.027       0.107\n",
            "no_shelter                       -0.2195      0.052     -4.204      0.000      -0.322      -0.117\n",
            "==============================================================================\n",
            "Omnibus:                      873.122   Durbin-Watson:                   1.887\n",
            "Prob(Omnibus):                  0.000   Jarque-Bera (JB):             1990.498\n",
            "Skew:                          -0.248   Prob(JB):                         0.00\n",
            "Kurtosis:                       4.398   Cond. No.                         15.1\n",
            "==============================================================================\n",
            "\n",
            "Notes:\n",
            "[1] Standard Errors assume that the covariance matrix of the errors is correctly specified.\n"
          ]
        }
      ],
      "source": [
        "# BUILDING / SUMMARISING OUR MODEL\n",
        "model_multiv = sm.OLS(y,X).fit()\n",
        "print(model_multiv.summary())"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Interestingly, we can see that the coefficient estimate for `post_second_edu` is now negative and significant, albeit at a lower level than before (p \\< 0.05).\n",
        "\n",
        "👉 **NOTE:** p-values are useful to machine learning scientists as they indicate which variables may yield a significant increase in model performance. However, p-hacking where researchers manipulate data to find results that support their hypothesis make it hard to tell whether or not a relationship held up after honest attempts at falsification. This can range from using a specific modelling approach that produces statistically significant (while failing to report others that do not) findings to outright manipulation of the data. For a recent egregious case of the latter, we recommend the [**Data Falsificada**](https://datacolada.org/archives/category/fake-data) series.\n",
        "\n",
        "## Predicting life satisfaction: the machine learning approach (30 minutes)\n",
        "\n",
        "Machine learning scientists take a different approach. Our aim, in this context, is to build a model that can be used to accurately predict how happy a person is using a mixture of features and, for some models, hyperparameters (which we will address in Lab 5).\n",
        "\n",
        "*Thus, rather than attempting to falsify the effects of causes, we are more concerned about the fit of the model in the aggregate when applied to unforeseen data.*\n",
        "\n",
        "To achieve this, we do the following:\n",
        "\n",
        "-   Split the data into training and test sets\n",
        "-   Build a model using the training set\n",
        "-   Evaluate the model on the test set\n",
        "\n",
        "Let's look at each of these in turn.\n",
        "\n",
        "### Split the data into training and test sets\n",
        "\n",
        "It is worth considering what a training and test set is and why we might split the data this way.\n",
        "\n",
        "A **training set** is data that we use to build (or \"train\") a model. In the case of multivariate linear regression, we are using the training data to estimate a series of coefficients. Here is a made-up multivariate linear model with three coefficients derived from (non-existent) data to illustrate things."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 13,
      "metadata": {},
      "outputs": [],
      "source": [
        "def sim_model_preds(x1,x2,x3):\n",
        "  y = 1.1*x1 + 2.2*x2 + 3.3*x3\n",
        "  return y"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "A **test set** is data that the model has not yet seen. We then apply the model to this data set and use an evaluation metric to find out how accurate our predictions are. For example, suppose we had a new observation where `x1 = 10`, `x2 = 20` and `x3 = 30` and `y = 150`. We can use the above model to develop a prediction."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 14,
      "metadata": {},
      "outputs": [
        {
          "data": {
            "text/plain": [
              "154.0"
            ]
          },
          "execution_count": 14,
          "metadata": {},
          "output_type": "execute_result"
        }
      ],
      "source": [
        "sim_model_preds(10,20,30)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "We get a prediction of 154 points!\n",
        "\n",
        "We can also calculate the amount of error we make by calculating residuals (actual value - predicted value)."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 17,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "❓**Question:** How many points are we off by?\n",
        "\n",
        "Why do we evaluate our models using different data? Because, as stated earlier, machine learning scientists care about the applicability of a model to *unforeseen* data. If we were to evaluate the model using the training data, we obviously cannot do this to begin with. Furthermore, we cannot ascertain whether the model we have built can generalise to other data sets or if the model has simply learned the idiosyncrasies of the data it was used to train on. We will discuss the concept of **overfitting** throughout this course.\n",
        "\n",
        "We can use `train_test_split` in `sklearn.model_selection` to split the data into training and test sets."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 16,
      "metadata": {},
      "outputs": [],
      "source": [
        "X_train, X_test, y_train, y_test = train_test_split(X, y, random_state = 123)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "👉 **NOTE:** Our data are purely cross-sectional, so we can use this approach. However, when working with more complex data structures (e.g. time series cross sectional), different approaches to splitting the data will need to be used.\n",
        "\n",
        "### Build a model using the training set\n",
        "\n",
        "We will now switch our focus from `statsmodels` to `scikit-learn`, the latter being Python's most comprehensive and popular machine learning library.\n",
        "\n",
        "Below, we will see how simple it is to run a model in this library. We have done all the data cleaning needed, so we can just get to it!"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 19,
      "metadata": {},
      "outputs": [
        {
          "data": {
            "text/html": [
              "<style>#sk-container-id-1 {\n",
              "  /* Definition of color scheme common for light and dark mode */\n",
              "  --sklearn-color-text: #000;\n",
              "  --sklearn-color-text-muted: #666;\n",
              "  --sklearn-color-line: gray;\n",
              "  /* Definition of color scheme for unfitted estimators */\n",
              "  --sklearn-color-unfitted-level-0: #fff5e6;\n",
              "  --sklearn-color-unfitted-level-1: #f6e4d2;\n",
              "  --sklearn-color-unfitted-level-2: #ffe0b3;\n",
              "  --sklearn-color-unfitted-level-3: chocolate;\n",
              "  /* Definition of color scheme for fitted estimators */\n",
              "  --sklearn-color-fitted-level-0: #f0f8ff;\n",
              "  --sklearn-color-fitted-level-1: #d4ebff;\n",
              "  --sklearn-color-fitted-level-2: #b3dbfd;\n",
              "  --sklearn-color-fitted-level-3: cornflowerblue;\n",
              "\n",
              "  /* Specific color for light theme */\n",
              "  --sklearn-color-text-on-default-background: var(--sg-text-color, var(--theme-code-foreground, var(--jp-content-font-color1, black)));\n",
              "  --sklearn-color-background: var(--sg-background-color, var(--theme-background, var(--jp-layout-color0, white)));\n",
              "  --sklearn-color-border-box: var(--sg-text-color, var(--theme-code-foreground, var(--jp-content-font-color1, black)));\n",
              "  --sklearn-color-icon: #696969;\n",
              "\n",
              "  @media (prefers-color-scheme: dark) {\n",
              "    /* Redefinition of color scheme for dark theme */\n",
              "    --sklearn-color-text-on-default-background: var(--sg-text-color, var(--theme-code-foreground, var(--jp-content-font-color1, white)));\n",
              "    --sklearn-color-background: var(--sg-background-color, var(--theme-background, var(--jp-layout-color0, #111)));\n",
              "    --sklearn-color-border-box: var(--sg-text-color, var(--theme-code-foreground, var(--jp-content-font-color1, white)));\n",
              "    --sklearn-color-icon: #878787;\n",
              "  }\n",
              "}\n",
              "\n",
              "#sk-container-id-1 {\n",
              "  color: var(--sklearn-color-text);\n",
              "}\n",
              "\n",
              "#sk-container-id-1 pre {\n",
              "  padding: 0;\n",
              "}\n",
              "\n",
              "#sk-container-id-1 input.sk-hidden--visually {\n",
              "  border: 0;\n",
              "  clip: rect(1px 1px 1px 1px);\n",
              "  clip: rect(1px, 1px, 1px, 1px);\n",
              "  height: 1px;\n",
              "  margin: -1px;\n",
              "  overflow: hidden;\n",
              "  padding: 0;\n",
              "  position: absolute;\n",
              "  width: 1px;\n",
              "}\n",
              "\n",
              "#sk-container-id-1 div.sk-dashed-wrapped {\n",
              "  border: 1px dashed var(--sklearn-color-line);\n",
              "  margin: 0 0.4em 0.5em 0.4em;\n",
              "  box-sizing: border-box;\n",
              "  padding-bottom: 0.4em;\n",
              "  background-color: var(--sklearn-color-background);\n",
              "}\n",
              "\n",
              "#sk-container-id-1 div.sk-container {\n",
              "  /* jupyter's `normalize.less` sets `[hidden] { display: none; }`\n",
              "     but bootstrap.min.css set `[hidden] { display: none !important; }`\n",
              "     so we also need the `!important` here to be able to override the\n",
              "     default hidden behavior on the sphinx rendered scikit-learn.org.\n",
              "     See: https://github.com/scikit-learn/scikit-learn/issues/21755 */\n",
              "  display: inline-block !important;\n",
              "  position: relative;\n",
              "}\n",
              "\n",
              "#sk-container-id-1 div.sk-text-repr-fallback {\n",
              "  display: none;\n",
              "}\n",
              "\n",
              "div.sk-parallel-item,\n",
              "div.sk-serial,\n",
              "div.sk-item {\n",
              "  /* draw centered vertical line to link estimators */\n",
              "  background-image: linear-gradient(var(--sklearn-color-text-on-default-background), var(--sklearn-color-text-on-default-background));\n",
              "  background-size: 2px 100%;\n",
              "  background-repeat: no-repeat;\n",
              "  background-position: center center;\n",
              "}\n",
              "\n",
              "/* Parallel-specific style estimator block */\n",
              "\n",
              "#sk-container-id-1 div.sk-parallel-item::after {\n",
              "  content: \"\";\n",
              "  width: 100%;\n",
              "  border-bottom: 2px solid var(--sklearn-color-text-on-default-background);\n",
              "  flex-grow: 1;\n",
              "}\n",
              "\n",
              "#sk-container-id-1 div.sk-parallel {\n",
              "  display: flex;\n",
              "  align-items: stretch;\n",
              "  justify-content: center;\n",
              "  background-color: var(--sklearn-color-background);\n",
              "  position: relative;\n",
              "}\n",
              "\n",
              "#sk-container-id-1 div.sk-parallel-item {\n",
              "  display: flex;\n",
              "  flex-direction: column;\n",
              "}\n",
              "\n",
              "#sk-container-id-1 div.sk-parallel-item:first-child::after {\n",
              "  align-self: flex-end;\n",
              "  width: 50%;\n",
              "}\n",
              "\n",
              "#sk-container-id-1 div.sk-parallel-item:last-child::after {\n",
              "  align-self: flex-start;\n",
              "  width: 50%;\n",
              "}\n",
              "\n",
              "#sk-container-id-1 div.sk-parallel-item:only-child::after {\n",
              "  width: 0;\n",
              "}\n",
              "\n",
              "/* Serial-specific style estimator block */\n",
              "\n",
              "#sk-container-id-1 div.sk-serial {\n",
              "  display: flex;\n",
              "  flex-direction: column;\n",
              "  align-items: center;\n",
              "  background-color: var(--sklearn-color-background);\n",
              "  padding-right: 1em;\n",
              "  padding-left: 1em;\n",
              "}\n",
              "\n",
              "\n",
              "/* Toggleable style: style used for estimator/Pipeline/ColumnTransformer box that is\n",
              "clickable and can be expanded/collapsed.\n",
              "- Pipeline and ColumnTransformer use this feature and define the default style\n",
              "- Estimators will overwrite some part of the style using the `sk-estimator` class\n",
              "*/\n",
              "\n",
              "/* Pipeline and ColumnTransformer style (default) */\n",
              "\n",
              "#sk-container-id-1 div.sk-toggleable {\n",
              "  /* Default theme specific background. It is overwritten whether we have a\n",
              "  specific estimator or a Pipeline/ColumnTransformer */\n",
              "  background-color: var(--sklearn-color-background);\n",
              "}\n",
              "\n",
              "/* Toggleable label */\n",
              "#sk-container-id-1 label.sk-toggleable__label {\n",
              "  cursor: pointer;\n",
              "  display: flex;\n",
              "  width: 100%;\n",
              "  margin-bottom: 0;\n",
              "  padding: 0.5em;\n",
              "  box-sizing: border-box;\n",
              "  text-align: center;\n",
              "  align-items: start;\n",
              "  justify-content: space-between;\n",
              "  gap: 0.5em;\n",
              "}\n",
              "\n",
              "#sk-container-id-1 label.sk-toggleable__label .caption {\n",
              "  font-size: 0.6rem;\n",
              "  font-weight: lighter;\n",
              "  color: var(--sklearn-color-text-muted);\n",
              "}\n",
              "\n",
              "#sk-container-id-1 label.sk-toggleable__label-arrow:before {\n",
              "  /* Arrow on the left of the label */\n",
              "  content: \"▸\";\n",
              "  float: left;\n",
              "  margin-right: 0.25em;\n",
              "  color: var(--sklearn-color-icon);\n",
              "}\n",
              "\n",
              "#sk-container-id-1 label.sk-toggleable__label-arrow:hover:before {\n",
              "  color: var(--sklearn-color-text);\n",
              "}\n",
              "\n",
              "/* Toggleable content - dropdown */\n",
              "\n",
              "#sk-container-id-1 div.sk-toggleable__content {\n",
              "  max-height: 0;\n",
              "  max-width: 0;\n",
              "  overflow: hidden;\n",
              "  text-align: left;\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-0);\n",
              "}\n",
              "\n",
              "#sk-container-id-1 div.sk-toggleable__content.fitted {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-0);\n",
              "}\n",
              "\n",
              "#sk-container-id-1 div.sk-toggleable__content pre {\n",
              "  margin: 0.2em;\n",
              "  border-radius: 0.25em;\n",
              "  color: var(--sklearn-color-text);\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-0);\n",
              "}\n",
              "\n",
              "#sk-container-id-1 div.sk-toggleable__content.fitted pre {\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-0);\n",
              "}\n",
              "\n",
              "#sk-container-id-1 input.sk-toggleable__control:checked~div.sk-toggleable__content {\n",
              "  /* Expand drop-down */\n",
              "  max-height: 200px;\n",
              "  max-width: 100%;\n",
              "  overflow: auto;\n",
              "}\n",
              "\n",
              "#sk-container-id-1 input.sk-toggleable__control:checked~label.sk-toggleable__label-arrow:before {\n",
              "  content: \"▾\";\n",
              "}\n",
              "\n",
              "/* Pipeline/ColumnTransformer-specific style */\n",
              "\n",
              "#sk-container-id-1 div.sk-label input.sk-toggleable__control:checked~label.sk-toggleable__label {\n",
              "  color: var(--sklearn-color-text);\n",
              "  background-color: var(--sklearn-color-unfitted-level-2);\n",
              "}\n",
              "\n",
              "#sk-container-id-1 div.sk-label.fitted input.sk-toggleable__control:checked~label.sk-toggleable__label {\n",
              "  background-color: var(--sklearn-color-fitted-level-2);\n",
              "}\n",
              "\n",
              "/* Estimator-specific style */\n",
              "\n",
              "/* Colorize estimator box */\n",
              "#sk-container-id-1 div.sk-estimator input.sk-toggleable__control:checked~label.sk-toggleable__label {\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-2);\n",
              "}\n",
              "\n",
              "#sk-container-id-1 div.sk-estimator.fitted input.sk-toggleable__control:checked~label.sk-toggleable__label {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-2);\n",
              "}\n",
              "\n",
              "#sk-container-id-1 div.sk-label label.sk-toggleable__label,\n",
              "#sk-container-id-1 div.sk-label label {\n",
              "  /* The background is the default theme color */\n",
              "  color: var(--sklearn-color-text-on-default-background);\n",
              "}\n",
              "\n",
              "/* On hover, darken the color of the background */\n",
              "#sk-container-id-1 div.sk-label:hover label.sk-toggleable__label {\n",
              "  color: var(--sklearn-color-text);\n",
              "  background-color: var(--sklearn-color-unfitted-level-2);\n",
              "}\n",
              "\n",
              "/* Label box, darken color on hover, fitted */\n",
              "#sk-container-id-1 div.sk-label.fitted:hover label.sk-toggleable__label.fitted {\n",
              "  color: var(--sklearn-color-text);\n",
              "  background-color: var(--sklearn-color-fitted-level-2);\n",
              "}\n",
              "\n",
              "/* Estimator label */\n",
              "\n",
              "#sk-container-id-1 div.sk-label label {\n",
              "  font-family: monospace;\n",
              "  font-weight: bold;\n",
              "  display: inline-block;\n",
              "  line-height: 1.2em;\n",
              "}\n",
              "\n",
              "#sk-container-id-1 div.sk-label-container {\n",
              "  text-align: center;\n",
              "}\n",
              "\n",
              "/* Estimator-specific */\n",
              "#sk-container-id-1 div.sk-estimator {\n",
              "  font-family: monospace;\n",
              "  border: 1px dotted var(--sklearn-color-border-box);\n",
              "  border-radius: 0.25em;\n",
              "  box-sizing: border-box;\n",
              "  margin-bottom: 0.5em;\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-0);\n",
              "}\n",
              "\n",
              "#sk-container-id-1 div.sk-estimator.fitted {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-0);\n",
              "}\n",
              "\n",
              "/* on hover */\n",
              "#sk-container-id-1 div.sk-estimator:hover {\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-2);\n",
              "}\n",
              "\n",
              "#sk-container-id-1 div.sk-estimator.fitted:hover {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-2);\n",
              "}\n",
              "\n",
              "/* Specification for estimator info (e.g. \"i\" and \"?\") */\n",
              "\n",
              "/* Common style for \"i\" and \"?\" */\n",
              "\n",
              ".sk-estimator-doc-link,\n",
              "a:link.sk-estimator-doc-link,\n",
              "a:visited.sk-estimator-doc-link {\n",
              "  float: right;\n",
              "  font-size: smaller;\n",
              "  line-height: 1em;\n",
              "  font-family: monospace;\n",
              "  background-color: var(--sklearn-color-background);\n",
              "  border-radius: 1em;\n",
              "  height: 1em;\n",
              "  width: 1em;\n",
              "  text-decoration: none !important;\n",
              "  margin-left: 0.5em;\n",
              "  text-align: center;\n",
              "  /* unfitted */\n",
              "  border: var(--sklearn-color-unfitted-level-1) 1pt solid;\n",
              "  color: var(--sklearn-color-unfitted-level-1);\n",
              "}\n",
              "\n",
              ".sk-estimator-doc-link.fitted,\n",
              "a:link.sk-estimator-doc-link.fitted,\n",
              "a:visited.sk-estimator-doc-link.fitted {\n",
              "  /* fitted */\n",
              "  border: var(--sklearn-color-fitted-level-1) 1pt solid;\n",
              "  color: var(--sklearn-color-fitted-level-1);\n",
              "}\n",
              "\n",
              "/* On hover */\n",
              "div.sk-estimator:hover .sk-estimator-doc-link:hover,\n",
              ".sk-estimator-doc-link:hover,\n",
              "div.sk-label-container:hover .sk-estimator-doc-link:hover,\n",
              ".sk-estimator-doc-link:hover {\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-3);\n",
              "  color: var(--sklearn-color-background);\n",
              "  text-decoration: none;\n",
              "}\n",
              "\n",
              "div.sk-estimator.fitted:hover .sk-estimator-doc-link.fitted:hover,\n",
              ".sk-estimator-doc-link.fitted:hover,\n",
              "div.sk-label-container:hover .sk-estimator-doc-link.fitted:hover,\n",
              ".sk-estimator-doc-link.fitted:hover {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-3);\n",
              "  color: var(--sklearn-color-background);\n",
              "  text-decoration: none;\n",
              "}\n",
              "\n",
              "/* Span, style for the box shown on hovering the info icon */\n",
              ".sk-estimator-doc-link span {\n",
              "  display: none;\n",
              "  z-index: 9999;\n",
              "  position: relative;\n",
              "  font-weight: normal;\n",
              "  right: .2ex;\n",
              "  padding: .5ex;\n",
              "  margin: .5ex;\n",
              "  width: min-content;\n",
              "  min-width: 20ex;\n",
              "  max-width: 50ex;\n",
              "  color: var(--sklearn-color-text);\n",
              "  box-shadow: 2pt 2pt 4pt #999;\n",
              "  /* unfitted */\n",
              "  background: var(--sklearn-color-unfitted-level-0);\n",
              "  border: .5pt solid var(--sklearn-color-unfitted-level-3);\n",
              "}\n",
              "\n",
              ".sk-estimator-doc-link.fitted span {\n",
              "  /* fitted */\n",
              "  background: var(--sklearn-color-fitted-level-0);\n",
              "  border: var(--sklearn-color-fitted-level-3);\n",
              "}\n",
              "\n",
              ".sk-estimator-doc-link:hover span {\n",
              "  display: block;\n",
              "}\n",
              "\n",
              "/* \"?\"-specific style due to the `<a>` HTML tag */\n",
              "\n",
              "#sk-container-id-1 a.estimator_doc_link {\n",
              "  float: right;\n",
              "  font-size: 1rem;\n",
              "  line-height: 1em;\n",
              "  font-family: monospace;\n",
              "  background-color: var(--sklearn-color-background);\n",
              "  border-radius: 1rem;\n",
              "  height: 1rem;\n",
              "  width: 1rem;\n",
              "  text-decoration: none;\n",
              "  /* unfitted */\n",
              "  color: var(--sklearn-color-unfitted-level-1);\n",
              "  border: var(--sklearn-color-unfitted-level-1) 1pt solid;\n",
              "}\n",
              "\n",
              "#sk-container-id-1 a.estimator_doc_link.fitted {\n",
              "  /* fitted */\n",
              "  border: var(--sklearn-color-fitted-level-1) 1pt solid;\n",
              "  color: var(--sklearn-color-fitted-level-1);\n",
              "}\n",
              "\n",
              "/* On hover */\n",
              "#sk-container-id-1 a.estimator_doc_link:hover {\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-3);\n",
              "  color: var(--sklearn-color-background);\n",
              "  text-decoration: none;\n",
              "}\n",
              "\n",
              "#sk-container-id-1 a.estimator_doc_link.fitted:hover {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-3);\n",
              "}\n",
              "</style><div id=\"sk-container-id-1\" class=\"sk-top-container\"><div class=\"sk-text-repr-fallback\"><pre>LinearRegression()</pre><b>In a Jupyter environment, please rerun this cell to show the HTML representation or trust the notebook. <br />On GitHub, the HTML representation is unable to render, please try loading this page with nbviewer.org.</b></div><div class=\"sk-container\" hidden><div class=\"sk-item\"><div class=\"sk-estimator fitted sk-toggleable\"><input class=\"sk-toggleable__control sk-hidden--visually\" id=\"sk-estimator-id-1\" type=\"checkbox\" checked><label for=\"sk-estimator-id-1\" class=\"sk-toggleable__label fitted sk-toggleable__label-arrow\"><div><div>LinearRegression</div></div><div><a class=\"sk-estimator-doc-link fitted\" rel=\"noreferrer\" target=\"_blank\" href=\"https://scikit-learn.org/1.6/modules/generated/sklearn.linear_model.LinearRegression.html\">?<span>Documentation for LinearRegression</span></a><span class=\"sk-estimator-doc-link fitted\">i<span>Fitted</span></span></div></label><div class=\"sk-toggleable__content fitted\"><pre>LinearRegression()</pre></div> </div></div></div></div>"
            ],
            "text/plain": [
              "LinearRegression()"
            ]
          },
          "execution_count": 19,
          "metadata": {},
          "output_type": "execute_result"
        }
      ],
      "source": [
        "# Create a model instance of a linear regression\n",
        "linear_model = LinearRegression()\n",
        "\n",
        "# Fit this instance to the training set\n",
        "linear_model.fit(X_train, y_train)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Evaluate the model using the test set\n",
        "\n",
        "Now that we have trained a model, we can then evaluate its performance on the test set. We will look at two evaluation metrics:\n",
        "\n",
        "-   **R-squared**: the proportion of variance in the outcome explained by the model.\n",
        "-   **Root mean squared error (RMSE)**: the amount of error a typical observation parameterised as the units used in the initial measurement."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 20,
      "metadata": {},
      "outputs": [
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "0.3 1.83\n"
          ]
        }
      ],
      "source": [
        "# Create predictions for the test set\n",
        "linear_preds = linear_model.predict(X_test)\n",
        "\n",
        "# Calculate performance metrics\n",
        "r2 = r2_score(y_test, linear_preds)\n",
        "rmse = root_mean_squared_error(y_test, linear_preds)\n",
        "\n",
        "# Print results\n",
        "print(np.round(r2, 2), np.round(rmse, 2))"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "🗣️ **CLASSROOM DISCUSSION:**\n",
        "\n",
        "How can we interpret these results?\n",
        "\n",
        "*We find that the model explains \\~ 33% of the test set variance in life satisfaction. We also find that our model predictions are off by \\~ 1.8 points.*\n",
        "\n",
        "## Graphically exploring where we make errors\n",
        "\n",
        "We are going to build some residual scatter plots which look at the relationship between the values fitted by the model for each observation and the residuals (actual - predicted values). Before we do this for our data, let's take a look at an example where there is a near perfect relationship between two variables. As this very rarely exists in the social world, we will rely upon simulated data.\n",
        "\n",
        "We translated and adapted this code from [here](https://www.r-bloggers.com/2021/05/how-to-generate-correlated-data-in-r/)."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 108,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Set a seed for reproducibility\n",
        "np.random.seed(123)\n",
        "\n",
        "# Create the variance covariance matrix\n",
        "sigma = [[1,0.99],[0.99,1]]\n",
        "\n",
        "# Create the mean vector\n",
        "mu = [10,5]\n",
        "\n",
        "# Generate a multivariate normal distribution using 1,000 samples\n",
        "v1, v2 = np.random.multivariate_normal(mu,sigma,1000).T\n",
        "\n",
        "# Combine to a data frame\n",
        "sim_data = pd.DataFrame({\"V1\":v1,\"V2\":v2})"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Plot the correlation"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 109,
      "metadata": {},
      "outputs": [
        {
          "data": {
            "text/html": [
              "   <div id=\"ZrwrUN\"></div>\n",
              "   <script type=\"text/javascript\" data-lets-plot-script=\"plot\">\n",
              "   \n",
              "   (function() {\n",
              "   // ----------\n",
              "   \n",
              "   var containerDiv = document.getElementById(\"ZrwrUN\");\n",
              "   var observer = new ResizeObserver(function(entries) {\n",
              "       for (let entry of entries) {\n",
              "           var width = containerDiv.clientWidth\n",
              "           if (entry.contentBoxSize && width > 0) {\n",
              "           \n",
              "               // Render plot\n",
              "               if (observer) {\n",
              "                   observer.disconnect();\n",
              "                   observer = null;\n",
              "               }\n",
              "\n",
              "               var plotSpec={\n",
              "\"data\":{\n",
              "\"V1\":[11.01239015284211,9.82424095538634,10.460377741270275,12.450933645734146,8.79852034951018,10.683883744820792,8.557520713658558,10.473583887064528,7.644962526048964,8.971151880237166,9.159066482995316,10.850347682808264,11.295837879085562,9.196188390129933,10.200653382634034,10.452869654713055,11.816587515682343,9.087137026785127,9.94849651321002,10.857279266308138,10.925515260182607,10.349347118006255,9.66309502465908,7.584425843258669,8.86545326710593,11.36429947446022,8.317084075396176,9.894779013444047,8.98743397317406,8.432482178068394,10.714569243440263,9.780295800080072,8.52916839854403,9.971105884779771,11.181193155214544,9.591504978453917,8.918289073469735,12.044266140184646,10.4112679967673,10.948979023058246,8.796615044149407,8.286119787268138,10.390755160419054,10.816366869903936,7.906483211496525,8.942288897516795,9.736130608957234,10.261261382885767,11.178265091966738,9.648360900867573,9.499410255784548,9.105790424871586,10.022150126241087,9.952561516722737,9.68843967361454,10.33872491061252,10.697932954436341,8.970086011470977,11.391026356139063,8.188685683606895,10.314135620968589,9.826648049922134,8.41444411485318,9.154355103754567,9.566715500031219,11.072609835111528,10.73179707065998,9.218082662702685,10.419879617533068,8.509967886812607,10.812162660084626,8.71234353460259,9.459840739690202,9.435655425911248,10.228935471735511,9.72675644750137,11.54106938605983,11.005227425238747,9.485054567534798,8.636314045797967,9.993450172224971,9.816494854626663,9.39954087643403,8.55758021971545,9.940374069437135,11.051677739521553,10.733804097398558,9.493841892663431,9.736605363200066,10.380741587506645,9.803564568477906,8.633520426726461,10.728079869099439,10.616705516247128,10.326247098025235,9.825370264838892,9.809835342715408,10.031944322664733,10.276457742029368,10.355818620785463,9.340742794409238,7.756136607788886,10.020904694627662,10.151027693208498,10.075258334207566,9.38091879669337,10.374321627900589,10.553428510162957,9.080048587984711,10.759130872757403,11.291553410902168,10.270161369965853,9.146111642095507,10.960471098210663,11.225483876667848,11.543127845240221,10.241333894625562,9.708930923491536,8.983846824121,9.42149615313089,9.69007857275434,9.819549737251206,9.587768949396587,9.523294435902889,11.435749359371721,8.082687999850066,10.800388842560626,9.264870915381803,9.402621542396641,9.761224863480546,10.017374953672242,10.16240695018214,9.037359748275534,9.34552467019162,8.754354237172246,11.304144368455818,8.63819854265624,11.364646121385208,8.83896446234288,10.911476662283588,11.331802568236203,11.54691882406335,10.402283183514786,11.560934178891772,11.306092015994913,11.0248191707221,9.108616872024307,11.869700820522345,8.099500788424491,9.781891613932988,9.295478550498137,10.614288406767558,10.050554788442922,10.64950294894744,9.66716017288272,8.778944369773505,10.95718129006365,8.63767498512114,10.295384635670858,9.96504811750612,10.910823423318206,10.472370647899455,9.508262766961222,11.775658114316,10.160374161110347,10.012860763914626,10.521817710579409,10.122637860041834,9.819660093101849,8.774218906443124,10.540727638929381,9.089717153422182,11.884228236228731,10.020894547810636,10.916005107107305,10.336363718277669,10.963587511828297,11.247487563754696,10.809770683985025,11.963550251300973,10.352786609838567,10.364923363041118,10.980141271787241,8.328754341968578,9.566850572521552,12.287545601563647,8.862128084351735,11.160389163250944,10.452464205797265,9.703127091243104,9.28339064030088,11.36494274724614,10.33748775315459,10.717983174007278,9.798294917158556,9.03113241885774,10.11363927279783,11.798146859686577,10.273426088257251,10.150749527623706,8.507220327733238,10.582315280867057,9.939567680057369,11.793770385169381,9.904038109983441,9.595074382189036,10.189677645788862,9.109079021243106,9.280271592893316,10.701413777200045,11.25682521810218,10.763529739439322,10.507973326700963,9.76868641608099,10.574237619101444,11.036873454298524,9.887925922243824,9.013105438094394,9.954870654597826,8.197370727426858,8.871429537329272,10.389815901840036,8.879274392409277,11.235162356782565,8.369891539183113,9.960213669583423,11.444568863233828,10.31834212443751,9.72118646159129,9.866193413901977,9.936777931189283,11.908052354632918,9.031560936808958,9.043573843728055,9.013140237243078,9.222504766465764,8.945722510627636,10.020775374979662,10.604809320060545,12.524788229452344,10.978352009551408,12.53584616098138,9.260995914835066,8.867585573769109,7.795330543546873,11.516125192233057,10.84108869879658,11.007434967957796,9.490204929252718,9.699242929752192,9.244924427281488,10.19904963099744,8.647762466930349,11.025587024726246,11.714106118213468,9.594308259811346,8.102679769424444,11.045584556013536,9.811575875119388,9.679735648529597,12.168330502295738,10.613452050562724,10.418078560047833,9.480640677364063,9.273494693563116,9.635656099057856,9.55468015760935,8.91535205625346,8.407942113472355,9.957184507425133,11.214827648975644,10.38504052960431,10.224050925387854,12.89336630892653,10.060574285207421,10.076212493282505,10.246092454376027,8.89310868212246,8.918181346464168,8.910241355407061,10.090408355641584,11.46886421267718,8.195751900747663,10.946800548359887,9.69485690440717,9.92259401735488,8.657398421144519,8.678778762835682,10.33548370949527,8.061458347856835,10.841585349600484,9.016971469911644,10.742408888292392,9.95806449540111,9.318517184772322,7.444089430564723,9.65432914522702,9.812230457457387,10.727637079599095,9.179651658398958,8.917930692288703,9.761753544001206,8.807066889359053,8.122436106970099,9.583825296231302,9.607908496420306,9.497289398494878,9.497213358035678,10.398216818821046,10.627683235001088,9.90076841887361,9.436441631054855,10.353528726911273,10.80420485075809,10.008431885069582,10.694696918065148,12.004794189797344,11.63760578984,9.10911465209368,9.84768541069774,8.54343456075257,10.494869917096397,10.077059137777313,10.784459992095822,10.361622839658736,9.197489577231735,9.814388050324716,9.230574450506898,9.640620375568187,12.178287789242649,11.215733641348981,10.720907048060102,9.968024337742607,10.053199069068842,8.234007401201916,10.338522481039707,8.81241556180948,9.91750028279836,11.485224017907111,9.102318455383603,9.071930840758721,9.90522725174964,10.716982986489588,9.563107357592667,9.860172727690408,11.581307334175584,10.034325767062295,11.45070566156268,8.988270240780878,9.128660052296398,10.028718142545145,8.363721760146532,10.195760653105673,8.417006646590185,9.172906003855381,11.152608017852332,10.607553989133542,10.188178694238353,10.862350968770112,9.688509763470964,9.099162357308497,9.310904312870916,10.251756369409295,9.55605471374927,9.697872167241975,9.977777279553463,10.397874010804694,12.340541772486072,10.555169379653941,10.536525160343528,10.286579103327632,10.06357327567344,10.074930780453181,9.05803773574713,10.641744810972392,8.910319742698249,8.52827375656095,11.759531338577355,10.05014082274631,10.841534907685356,8.870720664513888,11.693282599526965,12.86095313851229,12.16649533007599,12.308196116134965,7.64937149566211,10.061930975802936,10.24158111324374,11.592240493517261,8.83637492375814,10.341140681897237,11.32094106250166,10.725635237559938,10.694912957240499,11.146570174030042,11.301873332237262,10.961052758182538,10.069803630673913,9.188012503792518,10.545701607575225,10.262891947211905,10.906446077585674,11.227037620958404,9.54242090150017,10.08810858243892,9.039468898982943,11.296744617507322,8.960727076390034,10.561608916853384,11.232217639727056,9.366567904283897,10.305244442380468,9.980054638710603,11.10217818282943,11.150141540728601,10.404804758488986,10.502506468057922,9.146818959311478,8.491961321255566,10.260648644656783,9.963236281444926,10.820007274897073,10.426281416928617,10.522999535009674,10.680776474125393,7.244112865167569,10.550841803000313,10.028338259572319,10.363897771357045,11.715522255289276,9.787965888487278,10.520774849986616,8.405068398838933,10.190695625519737,8.87804705561452,9.56262798102298,10.42256163037921,10.72141495175403,9.944320734720982,8.958226537944725,10.543183471198558,11.693444461416682,10.327481646664827,8.142038354185695,11.83696877581698,9.78884489585107,9.298702730839086,9.461084436682626,11.077684592655222,8.99351069788035,9.420408331627225,8.206764899539829,10.299541560201241,10.262833116087425,10.619796227257908,10.06475766115494,10.688260866849678,10.835843496730323,10.474727674627246,8.920138458712334,9.624452255309476,9.064793005195575,9.882379013194871,9.8987317676786,11.37480635386141,10.905998621166207,10.342258116117621,10.373248407524981,9.468327597072232,10.255323210374497,13.209614414682783,11.198264041822842,8.984064699993525,10.690526955385733,9.928604661375614,10.746220440106113,9.482435666891112,9.907068070286595,9.664403594581449,10.664472219721365,11.946190695723141,10.345326457534838,9.733129227391661,10.374984640917976,9.372828308589554,8.394292831652201,10.126561707804731,10.034677977221047,9.922178734298553,8.469691420487871,10.270259969454873,9.088163465916091,10.549540706867145,11.171369847709991,11.792830856946177,10.38177598897333,11.451341279166979,10.373705793610343,8.997058679423107,9.609503385675469,10.706818042977142,9.354313974132987,9.65841329950892,10.282355072093203,10.895909856358264,9.40859456567377,10.662912599223121,9.748464357426233,7.906691084344271,10.01276956297443,10.359577268958098,10.354500148322987,9.792479987466551,9.045236076041208,10.720373454420447,8.83384218171542,9.666977229624617,10.147653341307123,9.439659932379918,9.2927353615998,10.93906300049278,10.803370659865926,11.391064332128625,10.788729527518395,10.478279526710882,10.990744446405373,9.552674883389834,8.185090432900543,9.73063284038196,11.860727723234405,8.698827050359379,9.794142565033491,9.504625906433452,9.631476462093117,8.75506012242543,9.137370189207122,8.344765666557423,10.552270003539178,10.083876484066082,9.598502290509703,9.852928250799305,9.638452662516226,12.339912736670804,9.399613503325687,9.032918340968203,9.074188333752552,10.23347933167434,10.457788042875325,9.297707087764982,9.283213006584042,10.531895861419835,9.228494303161808,9.787759227318297,9.59228431456088,8.467795921960871,9.432105401740529,9.787364198682177,11.11772087317757,11.146670161520746,11.639447008271674,8.314339332336154,10.348493495192583,9.385017441683988,10.070311898078733,10.30092047231244,10.371900170346667,8.71532145836847,10.334686282121574,10.13935488845216,10.909293862806543,9.692299930512704,12.745351555280479,9.583761791490899,10.561599813132819,8.91102534278296,8.194258507609684,9.466177985163217,11.534965879726242,8.919201214767906,10.164506991725903,9.515017540942749,10.03144080852411,11.026763391760342,8.078460483812833,11.43235395178063,9.355214567990792,11.055182189767972,10.418841242821234,10.20086202811355,11.221230441910535,9.001670531657247,9.685592482120626,10.604229581885614,11.157709420274982,8.63740242569217,10.811088653745632,8.067106619112055,9.138028136339225,9.69434881525765,10.97491080989896,8.878285389553488,8.600607655653437,10.699351724259438,9.633088374553449,9.675954011277264,10.869039047737065,9.046561913659977,8.9118004653814,11.06073596265928,7.792040426618601,11.069326811726656,9.82004984234424,9.399322530230396,10.20946784163378,11.773391301751692,10.875203251779315,10.692388413386524,8.841438040193651,11.02460561954069,9.531896094925475,8.745274401435365,9.871877477683288,9.825842987351722,10.978239100442156,10.45280366820455,8.842576875250806,9.940243232855234,10.167557089375402,8.834075003033579,9.547138506692555,9.830437944537747,10.314252649987228,9.143218223052008,9.8868221397845,11.054251308343787,10.343093234652184,9.820742324485083,9.64625838866547,8.595766857983516,10.17584316933347,10.91650656538032,10.248671707527624,11.680472261671182,11.055298441509581,9.64732256639263,9.484891497420321,10.208495756859579,9.341011278497565,8.340894077727175,9.600307146987326,11.410874394304999,10.457842011670802,10.592277121490676,9.129941629729307,10.929178312449114,9.88493004549246,10.328423287257769,7.921007634516926,8.731926097465845,9.706445169079936,10.299269681385207,9.60248376425516,10.63105105177457,7.942457820542904,8.71393886652617,9.179970556677652,10.356842235453744,11.498948592861533,9.195757311075191,8.697173148979234,10.123660783598618,9.094171922097463,10.296419067849268,9.394012640992775,10.862517378920746,9.998742309025662,11.179432419433189,10.150750143784762,10.009301674685986,9.749087978334607,10.51012745588203,10.832240892520804,7.854167418436759,9.785161342906921,10.554647959307243,9.173996158458285,10.093457165297732,9.318549935937977,9.50486165534742,9.111822299569125,9.61215150560874,10.543433705803444,9.28101862343368,10.074029137404166,9.813500010440347,11.27815857944568,9.82154049560028,10.140753152436147,10.207886445875836,9.343160684100699,8.995424621894909,9.20946777055341,9.359327536641526,9.79719434283576,9.49252211730627,9.20511027640948,9.267455770686436,9.550947848944354,9.242719422734632,9.37009050103839,9.625903845199462,9.371650808407228,9.211186972671522,8.937526417241319,8.23300777109067,9.454768087212853,10.073576149966874,9.229089618714228,10.259593088523996,10.303224760517367,11.607388542796988,10.460072180498658,10.560778916982686,9.524524136562334,11.013968754186356,9.143705941275881,9.346375750542098,7.642473780377213,9.942824139213887,11.707822861017664,9.662412591442974,9.52266564916259,9.917060393742164,10.329776014097819,8.404658147001799,8.733526715584349,9.456185597812343,9.678433665659398,8.218528196448382,8.758314349240612,10.239887655589333,9.851969555262903,10.188485070428651,9.351373044043681,10.64304052268491,10.162256911485017,8.885387077747138,10.777004755478973,11.439102223934976,9.391200859020762,8.330182926248058,10.543108348968797,11.710708578322615,11.422347718657646,10.773127661816794,10.438308239396372,9.671643215169613,10.36658887821122,11.296818933002193,9.739019717130123,10.321077851345821,8.604078129751155,10.259260319412434,9.842564974771847,11.730389552316598,9.04448447773923,9.437388864497008,10.924018909644994,10.851993432310518,10.489373150902374,11.96518713490054,8.863138994476433,11.123983053314488,11.086642587303206,10.544194180972646,8.715836764659697,9.71426760872946,10.272852560117801,9.950631548581233,9.45161043128862,10.292664279395387,9.74034255427072,10.799946228767139,10.749798248952981,10.340937534523983,9.790971447088022,9.016054206623508,10.16704992479012,7.933608824209243,10.040581979239988,12.849683989492096,9.381073256602214,10.437427395883438,11.177312633236246,8.671509908450425,9.83630891794801,11.09895845676904,9.553740829444036,8.281220591604372,10.686263012848201,10.47171162821039,10.484003780062148,10.540198275869786,9.40347109885592,8.47770028338346,10.539240213987812,11.18725307251164,10.004308005110849,10.429344940138316,9.066566414612401,11.260104864662905,10.290233807305475,11.837294175684109,10.61817361079558,9.211155056744673,10.147376358239345,9.061161526163776,10.653595577780456,9.844433833321114,13.871565953235772,10.81559463747813,11.511161036308204,10.245044290350057,9.38643166315589,8.216209912239185,10.517958312818463,10.546812510324902,9.539781414561949,9.95882002112079,9.50192847593551,10.058906467304322,11.527337669316879,12.86428204992842,9.056457180606106,9.350293972741948,11.759574632811741,8.566328112278823,10.192603854399144,10.964667237388031,10.591695653749703,9.713123534531324,11.29711244046765,9.848535912196063,10.158243487694152,8.728645122822682,12.136461235363099,9.693258007043662,13.497276517948068,10.337202480227115,11.675418631684966,9.619159774990207,10.798296730491064,10.434483855032784,9.086723648394138,9.105866981373788,10.188223055648745,9.948134506398308,10.234221217118645,10.129460981441353,11.274761315549966,11.808452246774253,11.29016188410362,10.850181482556726,8.571085569294748,9.835588310911309,9.909124619069749,10.366541071116597,9.897014894316149,11.412908943463009,9.758297134380975,8.619540467550218,11.47454887447953,10.858020998140502,8.551597894705196,10.72757849401902,10.073788504147123,10.53927820237893,11.011524080122301,10.048488369492778,9.841141321618617,10.847396120946863,9.524627673435463,9.16294163446017,9.350410798363184,11.773744830834426,11.33197910383537,9.924193349612553,11.03423939277673,9.760683326130517,9.707875878956495,11.21585969452586,8.459361409103181,8.834212838089831,10.189249964198572,11.131605509559064,9.462200348926435,10.357235478221769,8.680261701087115,10.43993500451495,8.888089899792393,10.344188463910752,10.188694542397622,9.118939892052943,9.508946471971845,9.32075909716107,8.995727722107231,9.234297245593398,9.960217466454964,10.773330881911154,10.382047529516266,9.614003119549658,10.039080038153111,11.806806220093954,9.463895592296906,10.230156172400022,10.713909302214224,10.064569074228089,10.21998087573199,8.342211130659663,9.071743974443672,10.047263536700324,8.30013206563989,11.355625129052349,10.080552381307147,10.907324375755415,9.662916949256672,10.836469524266125,10.473903471158392,9.667158271597486,11.747688451274866,9.202466078867337,10.013817052140205,9.468203044045307,11.285873435647346,9.663315879030216,10.111620170721677,9.777975273081362,10.823404112037164,9.109113348371357,9.585553373920266,11.520795172470423,10.949408001919615,10.215531951717756,7.883911455414671,10.044045248385896,8.818436200361324,9.848672872335973,10.03640289419466,9.822184389213177,9.050931383849726,9.85950617994576,10.103160740156282,8.805251043402023,10.937721311361495,9.758217032090085,9.552600957000937,9.213515978116524,9.16240615268324,10.06973131675021,12.437310757509739,9.492129784709663,10.088986502615034,8.772225818943927,11.096539032283742,10.980714343639747,8.978241144879622,10.302753597073401,11.57207781369404,8.439766123359805,9.342490507385252,9.15650296968437,9.825942875670647,10.722248440046386,11.012408524698678,11.00925641850865,12.844773104217815,9.752801732308429,8.797779289109776,9.050449843684145,10.079323319016533,10.511315690183933,9.66412558515428,11.257994278677163,10.549944381382568,9.400974386504851,11.471759493997666,9.889019269030532,9.401285773091988,8.34126283084878,10.967211912654781,9.570610715492558,10.163361383208695,10.604631966377829,10.201720615958557,10.465034321706947,11.066332005618541,10.551616902085762,8.097370894495961,11.142525313441332,11.60470625521636],\n",
              "\"V2\":[6.153436098535066,4.6112187140509535,5.693926136086399,7.390276240053996,3.6759447463161794,5.670489873985899,3.4671663267846906,5.4121573405831525,2.954220780552516,4.025766884568876,4.3698878282027955,6.01663502102457,6.205646196666219,3.994142427873907,5.078782836517417,5.057089387867842,6.717609927965889,4.062581233033744,5.045825902407423,5.897390227106924,5.681185897194245,5.430495521284497,4.661421940084928,2.642820440976412,4.181974533489808,6.217392636929709,3.2042209610159285,5.0460031280862525,4.2356123570719415,3.5837171397684155,5.826979799941984,4.592733540004594,3.643328884068435,4.938141689387206,6.2094101202992,4.473961913338833,3.763121130568397,7.191305755495364,5.393444721886308,5.721861591370979,3.699194258183959,3.400290405863541,5.237185370110343,5.64489020423535,2.9297387132969566,3.763058251915636,4.902705441803849,5.4070829881023785,5.9854410969358,4.594737255237096,4.219694662023521,4.473246090182737,5.026978552416345,4.689238386421636,4.4614004502992115,5.5144932079435325,5.768820266353544,4.009505921378822,6.344007251783495,2.902300965361176,5.236055748709175,4.9324617172626875,3.3762274834759936,4.225029032807628,4.48696461282007,5.9170402462051985,5.7772904225023165,4.263828065867125,5.675282359803761,3.459904710759317,5.830577835504026,3.7594036089941447,4.429848081167009,4.654086997004986,5.249202245626068,4.766881272113422,6.275640075266849,6.028978052424057,4.410005851570975,3.616065905328158,4.966019538590375,4.916122533611931,4.272484685698309,3.4027200929903976,4.901568533862735,6.041054084088951,5.74411473862218,4.702004140309158,4.6501650466075475,5.400537481116896,5.0099816623401905,3.582759221234986,5.3664577390193084,5.4783872689743385,5.381625472659547,4.821132148638498,4.791999596179451,4.575004141367435,5.260781082718169,5.324996364812013,4.256157927278253,2.853476493154868,4.991678231426335,5.021583726003422,5.114670135104088,4.4628993114568685,5.174057355205354,5.781427036793839,4.13232069117172,5.759646375517489,6.213497749383848,5.219017760331908,3.945473499203597,5.766060860945587,6.243028029792568,6.649740990368405,5.251062033904688,4.647530505364723,3.956362612132321,4.393337409917717,4.729628846857025,4.68176179976516,4.542661619416753,4.2188858937021845,6.487148323687288,3.2007831403619367,5.5610964951532615,4.2534747268305155,4.41892724406412,5.179637686636971,4.994854280205351,5.079882396291077,3.9873313306173506,4.3857736166571,3.8137766056415856,6.116550666861266,3.552113835302281,6.2699477508299,3.638116013885725,5.817175840040364,6.1643133897949625,6.48169707461798,5.305757712063645,6.738182648317119,6.345411920628883,6.119333448412957,3.984503568351135,6.9681000068477905,3.158288874036521,4.897823271133983,4.178241825322069,5.7007143280450245,5.2367509374395755,5.755655221396155,4.649279201143562,3.8751210082900025,6.047721884137085,3.619227320290981,5.200682227850666,5.062091549178693,5.7204197110983905,5.277248070183042,4.447919720996112,6.726357566787163,5.2238915286861545,5.277138593877452,5.51299322321166,5.082668170607834,4.896201757189621,3.552292576623292,5.477403196671754,4.039277379605439,6.896635197412071,5.0463152580277955,6.159113323125244,5.30973452117427,5.831923945780683,6.191865009672954,6.10059908443883,6.803980214893118,5.44803265980123,5.4605190228463725,5.988526304355479,3.1918946170377023,4.596114542941413,7.204252588097934,3.881230951991899,6.258079822044583,5.503383068076688,4.544919642112385,4.1409073792091755,6.24009187447688,5.352959695915156,5.823316531317936,4.700035611267074,4.195332280680331,5.0832360790110105,6.740515876186795,5.308127976824624,5.185260148173973,3.4322791572151914,5.397171346579543,5.077709896828432,6.699603989833438,5.024260812712714,4.641101889892469,5.2578249567522946,3.8673910864051098,4.2663048472059115,5.355910686170464,6.4969670237567465,5.610120268604832,5.402593595989258,4.983218891034925,5.596057527304412,6.251857454261818,4.734933594624608,4.182809318824965,4.723024912036879,3.241228674893594,3.7935214731196876,5.405436330356535,3.899496236121087,6.272213224311189,3.3791979528979414,4.781663021170007,6.354385576100989,5.3330585347137145,4.947730414165027,5.01672551410681,5.141683845002533,6.823378896166896,4.007293597020649,4.098030414223832,4.099841106806243,4.433569831440414,4.0862680029886,4.938489329180398,5.71100892345419,7.339925317218524,5.943333058343204,7.713856234718357,4.5416582491897675,4.01330321621952,2.714584876548835,6.483264894089284,5.804765279497243,6.0877453372772745,4.4863974999811305,4.679153923267087,4.254773116523088,5.330512727473877,3.8369070008573063,5.950004775727747,6.6803679297883365,4.575527337601955,3.0452848488962685,6.125983191947123,4.889307076463273,4.7142093329880925,7.0656424717413016,5.564242461969751,5.476459949205913,4.32060248272927,4.356334818174016,4.435135537559295,4.566360913859159,4.167698214256227,3.4238724050827534,5.046555490821572,6.009670071619294,5.557655040762918,5.436547954943708,7.6689014234216515,5.186183961846422,5.0626851524284415,4.9967810658693494,3.796551169452071,3.9082078290494406,3.9587853646161877,5.122406584634527,6.469716888302907,3.380758408787864,6.018315327286115,4.810424779282608,4.889571565959594,3.6880015173146456,3.8165777812115813,5.3118663042036784,3.0885380380160203,5.671873601684377,4.039640425345404,5.837809590022991,4.723237858895572,4.507058822091382,2.4490253985105306,4.881536244349854,4.851169677163058,5.5304426600556456,3.9995416282455,3.8064672500112726,4.709106370863705,3.849612995578223,2.9691809398309665,4.590520449236013,4.833605832957289,4.3350047988829745,4.438435902181227,5.5461233805822525,5.573981257070268,5.0345793116439586,4.40329035628918,5.189461616766072,5.711661390105314,5.289402608308199,5.652860370067514,7.089727768471661,6.509394670064983,3.8114328188001547,4.9609539675797905,3.4133137454853664,5.750055749195855,4.972935693644279,5.646919757799741,5.515520455323413,4.031422678416979,4.757514075166724,4.208461845821235,4.6181902120938165,7.184934938813154,6.321393554611746,5.635345337697524,5.049736742374737,5.285438783880993,3.476611296089102,5.128941906290294,3.7535354468993765,5.006358163895303,6.715616548240192,4.152137180284013,4.150695824230208,4.861701389696937,5.757601219625666,4.716294801343461,4.739511812184993,6.563267320219577,5.020453522388161,6.602657978966923,3.871840136231039,4.263410685038566,4.894580180439555,3.417592185930461,5.305057497123496,3.3466975843783064,4.44771035058559,6.0749318304162,5.585163150682872,5.2307175932236385,5.774293053693122,4.710417608771749,4.157053900116038,4.149266485324758,5.513156768905081,4.66175204001337,4.758023740985202,5.482874856475731,5.392454572218792,7.496991503563232,5.54657179473054,5.5169124301357755,5.28049207702187,5.041765351384108,5.259983234642672,4.183541032004227,5.81948611755969,3.832056551971533,3.4316510170854055,6.915310637469776,5.188298395345977,5.663093681364525,3.671702835688329,6.420963489882288,7.714001587358908,7.161459954152983,7.153442485100662,2.6981794338398393,4.832722032621797,5.248384691489148,6.610085136538345,3.787123788271221,5.052344637250527,6.470521738231197,5.594154157307487,5.686786019622336,6.164462309890759,6.4113334622585,6.147766894443068,4.950211448715735,3.9936292910058064,5.545690926499519,4.978835066156382,5.930238006456888,6.406137769954392,4.469696328013807,5.351373505553934,4.0932400061173055,6.527206460329904,3.9397393852083935,5.535035286989241,6.165638507659146,4.110931651692836,5.321331966345229,4.8133277757529145,6.256209428383421,6.2927863269398,5.557511022808641,5.337542511884439,4.144359469077059,3.6181374673308007,5.197056042573627,4.990136040140496,5.939075511184068,5.366456259483026,5.555670310271611,5.69986413013172,2.2365313426171283,5.314994930573286,4.8714693151975705,5.3853383228665,6.737790851290327,4.60452824453058,5.261879530229409,3.451817105730718,5.167582308690571,3.8647300355510708,4.7773946330719355,5.268039660914677,5.92345852882763,4.87328036265345,4.099359024894228,5.405730845520934,6.439750257280211,5.201164664449216,3.1503168428457524,7.037607394093188,4.888544050084907,4.340642162785828,4.494850294694772,6.12954835749242,3.9638042351176335,4.479747197603587,3.171020609225481,5.2830049149225,4.94143643719534,5.662206380084127,4.951714033874798,5.63796957595013,5.7632871835830395,5.601352234089782,4.052167328924563,4.524280328910332,4.020351261051627,5.071631975765108,4.933703010610632,6.443679868686093,6.049108499800073,5.02601608570395,5.340999453851638,4.563152368731475,5.398564404180367,8.108641122243538,6.020895708785455,4.101327102375449,5.857129374610612,4.868969523940282,5.6361561970285505,4.574381454750198,4.617123505769372,4.744362743479732,5.67058877015605,6.772246802890458,5.301521030652022,4.56527262947618,5.3057193201143,4.573794104163424,3.172481913052402,4.969226303896159,5.094487795323241,4.915247433383546,3.57165471809535,5.273683972949035,4.174400866300801,5.298483456627705,6.192014427693874,6.77127902035359,5.523091675604197,6.178147890233641,5.463474856569739,3.8684604194779735,4.4523233843894765,5.787088058925133,4.2129801931732676,4.3941675217898695,5.121948177907431,5.715445266554298,4.486924244467466,5.716027958379112,4.730298536117336,2.787772889877398,4.9586918706838965,5.382121530777222,5.232817428967869,4.95968693247279,4.073950902968479,5.74673858135172,4.002301782729022,4.615449340222281,5.256837123428825,4.191083669373572,4.393326723322559,5.830878333540497,5.7071669717353535,6.435342217432819,5.778775749961181,5.514367077868077,6.025430208361283,4.669714432043276,2.889222563164716,4.6303420695288295,6.979752267433286,3.6434425284004606,4.7850807455363515,4.629152466261662,4.680375714463452,3.731773367644163,4.021121774673157,3.449109934213123,5.25322773674867,4.91674339466371,4.803728553559445,4.827753465454181,4.678849700659538,7.373986354755221,4.362937057675212,4.014862429473653,4.395653940863235,4.95882036677808,5.2137103376814995,4.482877278303383,4.328511428444568,5.754726385449386,4.268724373682727,4.80354571977177,4.505546753080785,3.48934001642116,4.370789729312401,4.464675675956144,6.1132783824502495,6.491655406856924,6.688364761323838,3.3824036010176775,5.270408530018544,4.271842076613997,5.16806214232872,5.313296256637669,5.2347500779830325,3.9123307754401258,5.297128141576954,5.215553178153421,5.896774494033974,4.783098402040565,7.802401384288143,4.726276085443593,5.488293599833712,3.6899616013878935,3.5557165684331054,4.581308895407144,6.436740292703576,3.9703696318456063,5.197439660247167,4.473567549387695,5.0263061703386285,6.072526582917427,3.3233969172154225,6.513053392806948,4.297159621588667,5.868390838120168,5.468340009615675,5.114463042201126,6.2708292108262835,3.8705839692459683,4.419794544112371,5.70741118598386,6.072229158892371,3.533762636094506,5.8925259100769,3.033212170116396,4.32834863897346,4.656176659599331,5.773355542881225,3.903876594525137,3.740462815109022,5.531500597459507,4.543638393140479,4.572553512637652,5.849099936234821,4.089913201396065,3.737348496332925,6.048126311396894,2.887398883290575,6.058558925604854,4.739625532934176,4.377197705193041,5.387695065685814,6.637371485880213,5.755323045203407,5.761469345840762,3.97098583764494,6.009414434640776,4.565837637938641,3.6497677179312435,4.816266256667102,4.673263922718899,5.874985832447281,5.6717130757816205,3.812503670415022,4.714755928740448,5.190130018607375,3.8802921697147656,4.589248538398205,4.632906367540117,5.517176287083984,4.213113536209067,4.980081236329877,5.830797142148072,5.4191670430268015,4.9641377360159264,4.65260003287076,3.7559089763432487,5.019848361991202,5.893673926438119,5.483114704538466,6.671384838300503,6.071816720222317,4.590061487134633,4.456801406108654,5.149397366554555,4.360598343335753,3.520502718262799,4.294915980680531,6.55035128789876,5.4910539343855325,5.6655173511554535,3.958874434254839,5.693891583825197,4.649423829863269,5.401907400316461,2.9731098561459164,3.7685766924615374,4.676616166815348,5.465351832311838,4.650043249446383,5.5822547821574915,2.932911620135236,3.6403641639245228,4.220953843298001,5.448562880171104,6.580212334662585,4.2343756916901345,3.6875422903160917,5.1738709553229265,4.174759540490299,5.155672596218969,4.2823504005390465,5.769936853113288,4.886128983078725,6.086258098101943,5.214995669387383,4.856501259812886,4.794732501903509,5.397521258343968,5.757031273408802,2.7284335925924283,4.684171985235279,5.454785186359219,4.22761010173106,5.164582907552081,4.19824066030439,4.4033732092576505,4.216502172408765,4.582463437722899,5.506150977694019,4.184522441552534,4.8268770500787985,4.667054177226401,6.351515941550202,4.758837449689164,5.4619258278905995,5.179430202260134,4.355494371117814,4.034661807743497,4.2721641588189705,4.3532954957983305,4.758864135798886,4.302163500987155,4.346252374997119,4.252734130011184,4.441123040498785,4.087266905542366,4.445419632194446,4.562676669878495,4.111822319776325,4.238260381472902,3.965817616002113,3.099814403623914,4.329468130724108,5.178563488144701,4.523180799238355,5.441461142736152,5.2041025194452954,6.6303936433320425,5.5799292363745385,5.493377278324634,4.767940809112288,6.157880937156007,4.112459462491207,4.306172992784569,2.626621112189392,4.896750247997042,6.670607353136861,4.8746751042049254,4.4958829633262445,4.963353312310843,5.426258898289375,3.3533011598878915,3.85745797003517,4.472014270586355,4.647689628173713,3.1169306478362815,3.977478096852411,5.404763313737566,4.860788561146408,4.960269029587618,4.555522092357066,5.762872304902749,5.007727312271748,4.123828693095106,5.8075237239395445,6.43316149197074,4.63909971284055,3.562467269329459,5.508436772479653,6.609570316401421,6.474042442334862,5.617363030629586,5.270741333314496,4.711870165552712,5.263255760801805,6.375358217311718,4.804683839788274,5.290258813630511,3.519142603536694,5.36446254666542,4.553128886543107,6.63693389834858,4.139473395333202,4.526165285970784,6.009019804734981,5.98596177018765,5.529201092820015,6.774425435522987,3.9609663507671105,6.305793178265733,6.089465513689352,5.413603551496345,3.954129133005649,4.930076445168842,5.2881697089931965,4.787048208780317,4.551976197682696,5.053414534784817,4.526087288381736,5.818532006183922,5.785673544918817,5.1103505066347825,4.873926341015393,3.9642136523366673,4.9556096290499285,3.146449123495331,5.0229266864374695,7.996092225200929,4.496022213693665,5.37104013478609,6.287339824444832,3.830450949134142,4.790113391210252,5.999715946542608,4.51349890517294,3.44418834435179,5.514342434983699,5.622379529407539,5.561218141413671,5.588729016368074,4.542888973470315,3.6629946363253616,5.518448624930025,6.1539221395033366,4.9471794432475935,5.5753031186217115,4.1230107959238484,6.288520784980212,5.47630576925007,6.758264571362974,5.5521900784107014,3.878481339879802,5.267686429752155,3.98351019044632,5.7025575708887315,4.854248915975362,8.712159731305375,5.906529964230342,6.363145547175973,5.062950834165233,4.331078823921082,3.1271542372246968,5.63172306669615,5.760580552247184,4.496806301392494,5.3619709755401495,4.52320913961249,5.205583328603808,6.501713813374954,7.788846820044672,4.251934140470044,4.440799198900839,6.663638088368398,3.4922431191176653,5.07868018575522,5.898161125582601,5.727106856313781,4.857215467115225,6.470417686691947,5.1086590552720565,4.947044577657264,3.7093427710768103,7.156994864120847,4.845443330669446,8.659751197257954,5.609615071814562,6.752541198803559,4.559463737349009,5.745441201785209,5.257220952202978,3.9322631941188506,3.8528734289362023,5.1652694456206465,5.0833751666667055,5.169460218855373,5.2681963014341235,6.130416992038638,7.095148506776826,6.0235215822374855,5.763567230742576,3.505338890300214,4.582333802204683,4.675949617982571,5.259079086420616,4.98673383394703,6.252195108090724,4.8554940117251855,3.552804899046571,6.500941736848552,5.72728824006749,3.6390976651145386,5.724932283313721,5.0971365503300134,5.270568782784811,5.936288401521952,4.895428832945268,4.7872213944833835,5.667751948699943,4.5846247484332565,4.240622720293986,4.404444874944797,6.734271460351685,6.525614932528095,5.1037639433713196,6.0212471236687275,4.713886695917432,4.917485445169572,5.842801118724643,3.347403461915893,3.9264073011632705,5.243038840187871,6.143753201630218,4.352693164588991,5.401442877735995,3.679815539353595,5.419228687105752,3.797366355216143,5.254773287157123,5.218915664302178,3.8184795738561936,4.382856621595352,4.496069958587267,4.018066467026897,4.100882609856537,5.175103129710981,5.801228954774829,5.354455025045415,4.575887390442782,4.967721187829861,6.834113853589525,4.659364438159861,5.032914600332364,5.790332320592358,5.104066769850538,5.484411788265958,3.210463613843314,4.1147448899920365,4.925685295306345,3.464924480614382,6.5512803462089115,4.99601202758013,6.040942145659318,4.539293991094463,5.781665481379799,5.548542878698367,4.577263800242844,6.559032465737399,4.0617786325000935,4.7644689224498835,4.739665071654487,6.363612368965496,4.793124271563788,5.150203444043912,4.85898748145101,5.85062047687835,4.274292161883042,4.425470545685779,6.578583553477735,6.068814947030374,5.3585960504236745,2.875097698164746,5.187374795799102,4.059079935280311,4.7591719687694205,4.948208407914485,4.699396755015972,4.1495713874737215,4.863123249875851,5.288741650767778,3.8258585186030465,5.771289825976104,4.6679814475372545,4.49791394965519,4.242848923295019,3.846484242198132,5.383232501047577,7.556479776914584,4.673955371531469,5.257328041788994,3.800054726534268,6.225657371744415,5.961286362715297,4.132682841034503,5.3914223742730325,6.505273065975919,3.6297498830391914,4.3224320820409865,4.153218007263445,4.838442106445637,5.824909711389486,6.122129556207932,5.907139491460404,7.822699570786305,4.869767448103943,3.5870312850630888,3.984825490199443,5.064436125402609,5.3854632615889715,4.7788388174747265,6.255242394960158,5.704931125531675,4.614337509949048,6.6180169798367565,4.786464256253197,4.622756322448009,3.457542044417753,6.011033222486173,4.7462583907361005,5.165563565881834,5.69911378134727,5.228021960211474,5.471715011949112,6.089085169794888,5.650516894708481,3.2169711111283563,6.091713084587354,6.606625351773349]\n",
              "},\n",
              "\"mapping\":{\n",
              "\"x\":\"V1\",\n",
              "\"y\":\"V2\"\n",
              "},\n",
              "\"data_meta\":{\n",
              "\"series_annotations\":[{\n",
              "\"type\":\"float\",\n",
              "\"column\":\"V1\"\n",
              "},{\n",
              "\"type\":\"float\",\n",
              "\"column\":\"V2\"\n",
              "}]\n",
              "},\n",
              "\"theme\":{\n",
              "\"name\":\"minimal\",\n",
              "\"panel_grid_minor\":{\n",
              "\"blank\":true\n",
              "}\n",
              "},\n",
              "\"guides\":{\n",
              "\"x\":{\n",
              "\"title\":\"Variable 1\"\n",
              "},\n",
              "\"y\":{\n",
              "\"title\":\"Variable 2\"\n",
              "}\n",
              "},\n",
              "\"kind\":\"plot\",\n",
              "\"scales\":[],\n",
              "\"layers\":[{\n",
              "\"geom\":\"point\",\n",
              "\"mapping\":{\n",
              "},\n",
              "\"data_meta\":{\n",
              "},\n",
              "\"data\":{\n",
              "}\n",
              "}],\n",
              "\"metainfo_list\":[],\n",
              "\"spec_id\":\"3\"\n",
              "};\n",
              "               window.letsPlotCall(function() {\n",
              "       \n",
              "               var toolbar = null;\n",
              "               var plotContainer = containerDiv;               \n",
              "               \n",
              "                   var options = {\n",
              "                       sizing: {\n",
              "                           width_mode: \"min\",\n",
              "                           height_mode: \"scaled\",\n",
              "                           width: width\n",
              "                       }\n",
              "                   };\n",
              "                   var fig = LetsPlot.buildPlotFromProcessedSpecs(plotSpec, -1, -1, plotContainer, options);\n",
              "                   if (toolbar) {\n",
              "                     toolbar.bind(fig);\n",
              "                   }\n",
              "               });\n",
              "               \n",
              "               break;\n",
              "           }\n",
              "       }\n",
              "   });\n",
              "   \n",
              "   observer.observe(containerDiv);\n",
              "   \n",
              "   // ----------\n",
              "   })();\n",
              "   \n",
              "   </script>"
            ],
            "text/plain": [
              "<lets_plot.plot.core.PlotSpec at 0x1c431e83320>"
            ]
          },
          "execution_count": 109,
          "metadata": {},
          "output_type": "execute_result"
        }
      ],
      "source": [
        "(\n",
        "  ggplot(sim_data, aes(\"V1\", \"V2\")) +\n",
        "  geom_point() +\n",
        "  theme_minimal() +\n",
        "  theme(panel_grid_minor = element_blank()) +\n",
        "  labs(x = \"Variable 1\", y = \"Variable 2\")\n",
        ")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Residual plots"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 110,
      "metadata": {},
      "outputs": [
        {
          "data": {
            "text/html": [
              "   <div id=\"QXxcqt\"></div>\n",
              "   <script type=\"text/javascript\" data-lets-plot-script=\"plot\">\n",
              "   \n",
              "   (function() {\n",
              "   // ----------\n",
              "   \n",
              "   var containerDiv = document.getElementById(\"QXxcqt\");\n",
              "   var observer = new ResizeObserver(function(entries) {\n",
              "       for (let entry of entries) {\n",
              "           var width = containerDiv.clientWidth\n",
              "           if (entry.contentBoxSize && width > 0) {\n",
              "           \n",
              "               // Render plot\n",
              "               if (observer) {\n",
              "                   observer.disconnect();\n",
              "                   observer = null;\n",
              "               }\n",
              "\n",
              "               var plotSpec={\n",
              "\"data\":{\n",
              "\"predictions\":[6.0005800787994,4.825394968812448,5.454590763655871,7.4234273916919165,3.810866165852639,5.6756580571638775,3.57249611588621,5.4676528150640245,2.6698949978011246,3.981614086113016,4.167478322341406,5.840305680283576,6.280935061199129,4.2041951863617975,5.197700299374272,5.447164599041669,6.796002714546118,4.096333700025724,4.948294754209936,5.847161632158246,5.914653092101281,5.344771613424694,4.666007327318278,2.610018840517598,3.877068767592341,6.348649661068009,3.3346829205515487,4.895163374745238,3.997718522508576,3.448821895918895,5.706008741360229,4.781929306181377,3.544453157453381,4.970657431280987,6.167541244395407,4.5951984116503315,3.9293280739537444,7.021197100865921,5.406016863493548,5.937860838076922,3.8089816495858653,3.3040564887639103,5.385727846010157,5.806695637507861,2.9285621888825686,3.9530660320509305,4.738246008580008,5.257646995962915,6.164645129717815,4.651433969927169,4.504108549349221,4.114783567878016,5.021144720793423,4.9523154038138895,4.691075436722453,5.33426530664854,5.689553973588039,3.9805598472127954,6.375084921431453,3.2076855043178423,5.309944314708654,4.827775799151645,3.430980648919853,4.162818349719208,4.5706794152237205,6.060142694405846,5.723048592979362,4.225850566996309,5.414534521476616,3.525462145601322,5.80253730049977,3.7256296400675257,4.464970782625109,4.441049358704401,5.225673798041402,4.728974130111263,6.523490807660043,5.9934955048718255,4.489909498328652,3.6504297212741292,4.992757916603114,4.817733387007129,4.4053288608995835,3.5725549726629193,4.9402609357307075,6.039438992285913,5.725033720725586,4.498600943535953,4.738715582702747,5.375823532806368,4.804944185295401,3.647666583830752,5.719371950404405,5.609212821073898,5.321923641486299,4.826511956187474,4.811146538104106,5.03083205116069,5.272677545181663,5.351172504593622,4.3471723342282775,2.7798560412096416,5.019912878285144,5.148616083847628,5.073673456539693,4.386909969673569,5.369473622414718,5.5466262406058,4.089322604323987,5.750084151554841,6.276697341538904,5.266449873848952,4.154664833799474,5.949227521344387,6.211348702645566,6.525526806660414,5.237936939740196,4.711343103268062,3.9941705134172203,4.427044580032369,4.692696453569484,4.82075493731479,4.591503147008437,4.5277321253272405,6.419319944764769,3.10284437837939,5.790891948997167,4.272128305506378,4.408375913219155,4.763066455646056,5.0164216509024175,5.159871179913658,4.047099549427962,4.351902034192976,3.7671819570877005,6.289150917124539,3.6522936560833283,6.3489925256725055,3.8508689864490737,5.9007676716075,6.316507333827554,6.529276421502352,5.397130084987131,6.543138852534426,6.2910773135673,6.012873481661266,4.117579175210121,6.848536491553103,3.1194737199068108,4.783507707852037,4.302401975780828,5.606822085065591,5.049239455146049,5.6416523956978075,4.670028120054193,3.791503782787272,5.945973608824509,3.6517758111689123,5.291397924434888,4.964665761282804,5.900121560237798,5.466452813747102,4.512864469349081,6.755519900477479,5.1578605709311915,5.011956716197108,5.515360351553385,5.120536016728181,4.820864089054954,3.7868298797492166,5.5340639506261216,4.098885674485025,6.862905396284338,5.019902842181674,5.905246705839586,5.3319298776639705,5.952309931270554,6.233112303954767,5.800171423637071,6.941361916060364,5.348173576407046,5.360177903538492,5.968683070485481,3.346225850963682,4.570813013916332,7.261822100815802,3.873779866443323,6.1469642481425275,5.446763574044372,4.705602597569094,4.29044595872077,6.349285915000841,5.3330416480151115,5.70938542198037,4.799732031991265,4.04094016740553,5.111635614520514,6.777763267521266,5.269678970272674,5.148340953372383,3.52274456558208,5.5751978230970884,4.939463344995485,6.773434545423184,4.904321443789367,4.59872886904577,5.186844320563528,4.118036281844278,4.28736094374238,5.692996816533325,6.242348073650078,5.754435021493126,5.501667025979684,4.770446594042989,5.567208298001429,6.024796239005973,4.8883850587280815,4.023109881935936,4.954599346313302,3.2162757842243224,3.8829798297676703,5.384798835832939,3.8907390883690045,6.220921579303209,3.3869141935865112,4.9598840628008904,6.4280432175981925,5.314104920339975,4.723464919215546,4.866889677288869,4.936704035527113,6.886469565891252,4.041364009728289,4.053245841883049,4.023144301385498,4.230224420472215,3.956462180455489,5.0197849696646735,5.59744642620203,7.4964761361389645,5.966913331178402,7.507413412794124,4.268295585715405,3.879177808313706,2.818622325615003,6.498818783954077,5.831147722539349,5.995678960736324,4.495003663619432,4.701760816790486,4.252399456753867,5.19611404653037,3.661753226837609,6.013632957486564,6.694639508128972,4.597971105952871,3.1226180145325007,6.0334122924856075,4.8128680792783864,4.682466382658635,7.143907776718068,5.60599485447816,5.412753115620003,4.485543768825259,4.28065798825347,4.6388677922585,4.558775392363577,3.9264231030058756,3.4245495919244657,4.956887952319327,6.200808746771907,5.380075568453652,5.2208425477665195,7.861032600759035,5.059149627565823,5.074617204631117,5.242643578077962,3.9044224299165062,3.9292215223749043,3.9213681658404536,5.088658173647078,6.452073476566424,3.2146746208240415,5.935706133067251,4.697422648075198,4.92267488466355,3.6712840416016626,3.6924310989372984,5.331059470799287,3.081846366542095,5.831638954306252,4.026933730856674,5.733544623358554,4.957758338604611,4.325189231801733,2.4712136610501867,4.657337093810099,4.8135155193629515,5.718933992251694,4.1878388895770975,3.928973603127563,4.763589367648311,3.8193194530271892,3.142158787493713,4.587602523678706,4.611422948187309,4.502010833075403,4.501935622305913,5.393108089054218,5.6200707601685025,4.901087428806436,4.441826985436919,5.348907596120712,5.794666320113901,5.007576161563248,5.686353246131466,6.9821558352936925,6.618973887173183,4.1180715239208965,4.848583617891796,3.5585636592658094,5.4887065897900875,5.075454611251331,5.775136900788128,5.356913392698907,4.205482176001331,4.8156495703866184,4.238206054889634,4.643777902747393,7.1537564181016045,6.201704853716377,5.712277393171963,4.967609507500411,5.051854883599701,3.252512709344579,5.334065085817436,3.8246097652500985,4.917636728666082,6.468254777323661,4.111349481608931,4.081293431146873,4.905497610638019,5.708396147669955,4.567110641809087,4.860934683755321,6.563289713035999,5.03318751123445,6.434113056330991,3.9985456654646967,4.137403661110531,5.027641072412571,3.380811734003805,5.19286095520669,3.4335152204598955,4.181166837957903,6.13926800407598,5.600161147748226,5.185361724315076,5.852177996456902,4.691144761817153,4.108227820228807,4.317659425173131,5.24824569322106,4.560134950499824,4.700405010891037,4.977256033390619,5.3927690214714055,7.314240021972271,5.548348115189073,5.529907325961721,5.282688470674849,5.062115895563313,5.073349476777137,4.067551916044167,5.6339789080171165,3.921445697835269,3.543568277053229,6.739569086589174,5.0488300062958995,5.831589062770991,3.8822786915778718,6.674043198212409,7.828973095673437,7.142092628358311,7.282247294292976,2.6742558604359266,5.060491515091499,5.23818146091617,6.574103578913254,3.8483077027373653,5.3366547190632545,6.305764339733374,5.716953992633801,5.68656692806854,6.133296040141168,6.286904660548343,5.949802834730523,5.068278270155805,4.196108508438478,5.5389836474980765,5.259259768938781,5.895791976430349,6.212885493781645,4.546649898974469,5.086383493640641,4.049185686851926,6.281831906051767,3.9713030289815396,5.554717389559832,6.218008992530294,4.372715662165094,5.301150149203709,4.979508543914797,6.089388417114243,6.136828439081837,5.399624146650211,5.496259817112765,4.155364433363255,3.507652052584424,5.257040943419545,4.962873694457013,5.810296321455144,5.420866469806745,5.516529280906736,5.672584690400049,2.2734190722019987,5.544067757422084,5.027265334404113,5.359163502250649,6.69604019350481,4.7895157050172585,5.51432886978067,3.4217072328547626,5.187851192953422,3.889525143243908,4.5666364957944445,5.417187270448151,5.7127797550467925,4.944164538313288,3.9688297743143917,5.536492986913435,6.67420329400095,5.323144719766628,3.1615471514293034,6.816161591465182,4.790385121389974,4.305590976570215,4.466200909439117,6.065162080451735,4.003728943091794,4.425968627924529,3.225567455010667,5.295509492091464,5.2592015797301865,5.612269808838496,5.063287358023857,5.679987419708546,5.825959751718564,5.468784122576222,3.9311572802557677,4.62778619545575,4.074233477658451,4.882898672898138,4.899072999841,6.35904189821394,5.895349402280022,5.337759960778687,5.368412112119673,4.473365038537835,5.251773615809379,8.173830072442017,6.184425867807439,3.9943860120109704,5.682228782587204,4.928619945578227,5.737314586715867,4.487319172932157,4.907318343707107,4.667301619223233,5.656458334344697,6.9241917733924705,5.3407948230279665,4.735277375572636,5.3701294011901695,4.378907758189953,3.4110492396190413,5.124417050755461,5.033535878371369,4.92226413282761,3.485625142294772,5.266547397503048,4.097348939513829,5.542780857758611,6.157825120582933,6.7725052685439096,5.376846647751865,6.434741738609123,5.368864507591709,4.007238212035867,4.613000435348651,5.698342115495153,4.360595436621864,4.661376685226106,5.278510528526295,5.885370717611595,4.414283763876226,5.65491573111718,4.750445181530786,2.9287677935961547,5.011866510365076,5.354890141550008,5.349868418196774,4.793980549934721,4.054889937311417,5.711749621658546,3.845802595851449,4.669847172919475,5.145278550043924,4.4450101713573105,4.299688718258869,5.928053010832512,5.793841231213067,6.375122483058869,5.779359850621292,5.472297219829079,5.979170552155715,4.556791998014397,3.2041294819716395,4.73280822704411,6.839661301176479,3.7122606363632684,4.795624986849758,4.509267291192931,4.6347337599866005,3.7678801403130713,4.146018760313012,3.3620624734082973,5.545480374634588,5.0821975723755255,4.6021193749060405,4.853769252519665,4.6416338419620615,7.3136178496682085,4.405400695347515,4.042706603031887,4.083526292270655,5.2301680791917935,5.452029321877667,4.304606197118229,4.290270263194058,5.525328538065006,4.236148604367624,4.789311298761612,4.595969244295372,3.4837503258578266,4.4375380694198086,4.788920580346636,6.104761518661913,6.133394936652238,6.620795015773673,3.331968125825613,5.343927304627179,4.390963893687836,5.068780991773018,5.2968733586702825,5.3670785857746,3.7285750712270396,5.330270744284242,5.137070642946094,5.898608689007501,4.694893573735179,7.714632859955692,4.587539711848764,5.554708385171429,3.922143599003684,3.2131975223393834,4.47123888139096,6.5174538976681795,3.930230262731138,5.161948307565529,4.519545540902564,5.03033403094404,6.014796488990276,3.0986629894512783,6.415961585013376,4.361486203954877,6.042905204910641,5.413507476638689,5.197906668289319,6.207141677800708,4.011799743606671,4.688259311394004,5.596873013643803,6.144313744274253,3.6515062256798174,5.801475012817113,3.0874330085989303,4.1466695284778705,4.696920102768313,5.963509679010215,3.889760876682841,3.61511292930971,5.690957262984837,4.6363280846101755,4.678726007460108,5.858793100715872,4.0562013085021515,3.922910264114474,6.048398379690449,2.815368107683029,6.056895492618477,4.821249585680517,4.405112897106584,5.206418582390642,6.753277821470558,5.86489004613391,5.684069929919611,3.853315574693079,6.012662260571229,4.5362399302613134,3.7582011929025647,4.8725117213337725,4.826979520748745,5.966801654038646,5.447099332522599,3.8544419837373027,4.940131526727846,5.164965125107781,3.8460328768186667,4.55131602961594,4.831524341616031,5.310060066804802,4.151802985317689,4.887293319817145,6.0419844804837695,5.338585967200265,4.821934512037068,4.649354398574708,3.6103249503636885,5.173160794311391,5.90574269262053,5.245194688582545,6.661372636018491,6.043020188210754,4.6504069648871,4.489748207498488,5.205457104190149,4.347437888843427,3.3582331281080258,4.603904538303624,6.394716394452692,5.452082701810916,5.58505096847143,4.1386712550454305,5.918276176213883,4.885421870471003,5.324076085967231,2.942928133581497,3.744998534429916,4.708884471323646,5.295240579789027,4.6060574061494926,5.623401830150147,2.9641442731691505,3.7272075650672587,4.188154308305254,5.352184950437213,6.481829601227824,4.203768810880128,3.7106247809677555,5.1215477790011015,4.103291836461009,5.292421069757008,4.399860944926592,5.852342590879552,4.997992310045877,6.165799720286062,5.148341562810407,5.008436460764905,4.7510619980108775,5.503797651109927,5.822396456104455,2.8768172218761974,4.7867417602859295,5.547832384143851,4.1822450977081616,5.091673717249481,4.325221625614139,4.509500467812487,4.120749627818158,4.615619661218352,5.536740491168201,4.288099823312068,5.072457671700809,4.814771220048613,6.2634486632482,4.8227239741785315,5.138453650246296,5.204854441496794,4.349563841935458,4.005621983972832,4.2173296733243255,4.365554295341493,4.798643466212363,4.497295568460512,4.213019724501596,4.274684957038003,4.555083807520216,4.250218511367153,4.376199823298419,4.629221946814393,4.377743105901974,4.2190301169509254,3.948355515965636,3.251523986358656,4.459953478676848,5.072009626882483,4.236737424336065,5.255996904629919,5.299152504415066,6.589086345384841,5.4542885369071,5.553896445951549,4.528948408524517,6.0021414557698485,4.152285381969383,4.3527438282685225,2.6674334072346806,4.9426842724010775,6.688424808565534,4.665332340241028,4.52711019938045,4.917201639310305,5.325414052986618,3.4213014573415723,3.7465816879460476,4.4613555225896295,4.681178605933463,3.237202400689954,3.771098859734767,5.236506480851281,4.852821017464839,5.1856647575853625,4.357686595494954,5.635260482010221,5.1597227783129505,3.8967850772115433,5.767763009460078,6.422636225682163,4.397079842761043,3.3476388477297476,5.536418684354436,6.691279039343682,6.406064531513385,5.763928219391984,5.432762065675368,4.674462247143854,5.361825246004027,6.28190541067829,4.741103593227788,5.316810797290886,3.618545536504975,5.2556677664157885,4.843519051867901,6.710745271057087,4.054146539819699,4.442763883462254,5.913173068485653,5.8419334727476535,5.483269799698297,6.94298093935905,3.874779746364652,6.110955371141653,6.074022332975741,5.537492668670213,3.7290847549282358,4.716621559072875,5.269111699991228,4.950406493865962,4.4568302764455385,5.2887072504053645,4.742411996679051,5.790454164634147,5.740853357513526,5.336453788263745,4.792488469483498,4.026026475976174,5.164463494277063,2.955391829661285,5.0393754608723285,7.817826906155222,4.387062744245444,5.431890833188028,6.1637030634181045,3.685241556194258,4.837331255888786,6.086203822004041,4.557846313359331,3.2992107490139713,5.6780113646142105,5.465800984728123,5.477959014929059,5.533540363532223,4.409216200030926,3.4935466191959694,5.532592755217921,6.173535040957782,5.003497277923673,5.423896566420744,4.075987537119188,6.245591935458373,5.28630329758831,6.816483440841894,5.610664897006018,4.218998549263901,5.14500458918306,4.0706416222799024,5.645700369197737,4.845367519000206,8.82855895102579,5.805931831034567,6.493908792734933,5.241606850738682,4.392362684361229,3.234909411653561,5.511543064303693,5.540082429008731,4.544039211903817,4.958505620661343,4.50659929279408,5.057500007492781,6.509908919892554,7.8322656848003565,4.065988606595134,4.3566192981141825,6.739611908433188,3.581207415560039,5.189738600839846,5.953377875770611,5.5844758452187335,4.715489968052784,6.282195715636467,4.849424839431137,5.1557531456361865,3.7417533590301044,7.112386240495484,4.695841196538629,8.458353448116373,5.332759487746545,6.656374147003895,4.622551462242596,5.788822664299726,5.4289794097862085,4.09592483207301,4.114859289059877,5.185405601690967,4.947936697312915,5.230901869804395,5.12728468994996,6.2600884674466215,6.787956210868803,6.275320997913623,5.840141293469781,3.585912963195378,4.836618511511652,4.909352453154893,5.361777960540321,4.8973748595606335,6.396728744316791,4.760170671470753,3.633839162243188,6.457696112107261,5.847895270834989,3.5666379217654356,5.718876045908329,5.072219664032032,5.53263032911158,5.999723455938682,5.047195583226257,4.842110932423774,5.83738632338955,4.529050815690129,4.171311191403804,4.356734849032227,6.753627493140467,6.316681943218333,4.924256766349085,6.0221909181541555,4.76253082709737,4.7102995704940565,6.20182953150672,3.4754078433298954,3.8461692079343948,5.1864213054760295,6.118494657537111,4.467304645784505,5.35257390246999,3.6938978566103646,5.434371080897439,3.8994584083296644,5.3396692462358,5.185871943971904,4.127789559102755,4.513540714344092,4.327406682322925,4.005921777009865,4.241888229958686,4.959887818244155,5.764129222119125,5.377115225472155,4.617451071901009,5.0378899077068215,6.786328144251016,4.468981392081222,5.226881179444138,5.705356000859316,5.063100828799373,5.2168169070787,3.3595358104636084,4.081108603760319,5.045984114781891,3.31791587690703,6.340069962841522,5.078909739437766,5.8966606913471455,4.6658311949253,5.826578948566983,5.467968912124311,4.670026239514188,6.727855418939573,4.210404378225183,5.012902570207668,4.4732418445290705,6.2710793414099175,4.66622577191176,5.109638543192672,4.779634098826907,5.8136560951575,4.118070234423698,4.589311746033374,6.503437809284094,5.938285136280767,5.212416525991722,2.906236716870759,5.042800941730173,3.8305647116439445,4.849560305176878,5.03524197449776,4.823360842164398,4.0605231026981325,4.860275408820588,5.101271414838161,3.8175234200581816,5.926725961087671,4.760091443188875,4.556718878248249,4.221333710241993,4.170781552354882,5.068206745260982,7.409953144997459,4.496907516668125,5.087251834688159,3.7848585391490843,6.083810796249406,5.969249889458604,3.9886259986458956,5.298686481911407,6.554160897340584,3.4560263649087783,4.3489009776262115,4.164942779937533,4.827078319170116,5.7136041490139995,6.00059825019768,5.997480537146531,7.812969604341992,4.754735230616291,3.8101331912817518,4.060046816694943,5.077694087643977,5.504972920395774,4.6670266432017495,6.243504378416629,5.54318012771631,4.4067467297108855,6.454937167005591,4.889466475831711,4.407054718708762,3.358597857704863,5.955894785675225,4.574532129361103,5.160815198967569,5.5972710076539425,5.198755887943912,5.459196535392577,6.053933363322685,5.544834399276436,3.1173670655943564,6.1292953136785275,6.586433324676493],\n",
              "\"residuals\":[0.15285601973566543,-0.2141762547614947,0.23933537243052871,-0.03315115163792015,-0.13492141953645964,-0.005168183177978847,-0.10532978910151947,-0.055495474480872,0.2843257827513912,0.044152798455860065,0.2024095058613895,0.1763293407409936,-0.07528886453290973,-0.2100527584878904,-0.11891746285685567,-0.3900752111738264,-0.07839278658022852,-0.03375246699197998,0.0975311481974872,0.050228594948678484,-0.2334671949070355,0.0857239078598031,-0.004585387233349891,0.032801600458814306,0.304905765897467,-0.13125702413829998,-0.13046195953562023,0.15083975334101485,0.23789383456336566,0.13489524384952034,0.12097105858175539,-0.18919576617678313,0.09887572661505395,-0.03251574189378115,0.04186887590379307,-0.12123649831149841,-0.16620694338534747,0.17010865462944302,-0.012572141607240184,-0.21599924670594284,-0.1097873914019063,0.0962339170996307,-0.14854247589981462,-0.1618054332725114,0.0011765244143879805,-0.19000778013529462,0.1644594332238407,0.14943599213946346,-0.1792040327820157,-0.05669671469007298,-0.28441388732569983,0.35846252230472153,0.005833831622922503,-0.26307701739225386,-0.2296749864232419,0.18022790129499278,0.07926629276550567,0.02894607416602657,-0.031077669647958217,-0.30538453895666606,-0.07388856599947857,0.10468591811104222,-0.05475316544385933,0.06221068308841993,-0.0837148024036507,-0.1431024482006471,0.05424182952295453,0.03797749887081636,0.2607478383271449,-0.06555743484200516,0.02804053500425585,0.03377396892661899,-0.035122701458099925,0.21303763830058475,0.02352844758466599,0.037907142002159055,-0.24785073239319377,0.035482547552231125,-0.07990364675767747,-0.034363815945971155,-0.02673837801273926,0.09838914660480214,-0.13284417520127434,-0.16983487967252175,-0.03869240186797285,0.0016150918030382044,0.019081017896594332,0.20340319677320462,-0.08855053609519992,0.0247139483105272,0.20503747704478936,-0.06490736259576568,-0.35291421138509627,-0.1308255520995596,0.05970183117324801,-0.0053798075489757125,-0.019146941924654826,-0.45582790979325516,-0.011896462463494295,-0.02617613978160893,-0.09101440695002427,0.0736204519452266,-0.02823464685880861,-0.12703235784420563,0.040996678564395594,0.0759893417832993,-0.19541626720936378,0.23480079618803895,0.04299808684773243,0.009562223962648275,-0.06319959215505655,-0.04743211351704435,-0.20919133459587735,-0.18316666039879959,0.03167932714700239,0.12421418370799042,0.013125094164491813,-0.06381259790333882,-0.03780790128489908,-0.03370717011465185,0.03693239328754139,-0.1389931375496305,-0.04884152759168359,-0.308846231625056,0.0678283789225187,0.09793876198254647,-0.22979545384390576,-0.01865357867586237,0.010551330844965179,0.41657123099091464,-0.021567370697066934,-0.07998878362258122,-0.05976821881061145,0.033871582464123584,0.046594648553885065,-0.1726002502632733,-0.10017982078104737,-0.07904477484260575,-0.21275297256334857,-0.08359183156713623,-0.15219394403259123,-0.0475793468843726,-0.09137237292348566,0.19504379578269315,0.05433460706158311,0.10645996675169034,-0.133075606858986,0.11956351529468723,0.03881515412971037,0.11431556328194592,-0.12416015045875906,0.09389224297943333,0.18751148229352665,0.11400282569834719,-0.020748918910630465,0.08361722550273054,0.10174827531257513,-0.032548490877931435,-0.09071569658422174,0.0974257878958893,-0.17970184913940734,-0.18920474356406025,-0.0649447483529686,-0.029162333690315734,0.06603095775496293,0.2651818776803436,-0.0023671283417252553,-0.03786784612034744,0.07533766813466691,-0.2345373031259248,-0.056660753954367316,-0.05960829487958641,0.03372980112773316,0.02641241584612164,0.2538666172856585,-0.022195356489700835,-0.12038598548987078,-0.041247294281813396,0.3004276608017591,-0.13738170116724646,0.09985908339418437,0.10034111930788026,0.019843233869997867,-0.15433123392597992,0.025301529025080427,-0.05756951271786814,0.007451085548575875,0.11111557390205551,0.05661949403231592,-0.16068295545670885,-0.14953857951159488,-0.10919404052396153,0.019918047900044478,0.11393110933756567,-0.09969642072419127,0.15439211327480074,-0.028399535509503693,-0.037247391334470414,0.03844900655195005,0.036919194801590116,-0.09046540836688877,-0.17802647651754544,0.13824655183294787,-0.07383055558974583,0.11993936892334656,0.04237302084669903,0.07098063618876616,-0.25064519543916797,-0.02105609653646834,-0.33708613036286117,0.2546189501066687,-0.14431475288829443,-0.09907342999042523,0.21277229699193612,0.02884922930298295,0.2270612152558451,-0.15345146410347343,0.15969943688902966,-0.231574434276423,0.024952890669271532,-0.08945835664798274,0.02063749452359609,0.008757147752082695,0.051291645007980335,-0.007716240688569798,-0.17822104163088337,-0.07365764149720366,0.018953614373739214,0.22426549494948045,0.14983583681794066,0.20497980947542072,-0.06309066972435584,-0.03407041270764033,0.0447845723407827,0.07669680542074442,0.2033454109681987,0.12980582253311113,-0.08129564048427529,0.11356249725216028,-0.1565508189204401,-0.023580272835197924,0.2064428219242327,0.2733626634743622,0.13412540790581406,-0.10403744906616774,-0.015553889864793113,-0.02638244304210602,0.0920663765409504,-0.008606163638301467,-0.022606893523398952,0.0023736597692209926,0.13439868094350693,0.17515377401969712,-0.06362818175881735,-0.014271578340635749,-0.022443768350915327,-0.07733316563623216,0.09257089946151531,0.07643899718488623,0.03174295032945729,-0.07826530497676654,-0.041752392508409564,0.06370683358591034,-0.16494128609598935,0.07567682992054614,-0.2037322546992053,0.007585521495582448,0.24127511125035106,-6.771868417123628E-4,0.08966753850224496,-0.1911386751526125,0.17757947230926607,0.21570540717718867,-0.19213117733738372,0.12703433428059885,-0.011932052202675791,-0.24586251220861222,-0.10787126046443518,-0.021013693325463656,0.03741719877573413,0.03374841098744952,0.01764341173648276,0.16608378796382262,0.08260919421886381,0.1130021312074092,-0.03310331870395622,0.0167174757129831,0.12414668227428294,-0.01919316659560888,0.0066916714739253536,-0.15976535262187497,0.012706694488729298,0.10426496666443708,-0.23452047970903944,0.181869590289649,-0.02218826253965611,0.22419915053975448,0.03765415780010617,-0.1884913321960484,-0.18829726133159763,-0.12250635311629043,-0.05448299678460611,0.03029354255103378,-0.1729778476627466,0.0029179255573072993,0.22218288476998005,-0.1670060341924282,-0.0634997201246863,0.1530152915280345,-0.04608950309823445,0.13349188283752245,-0.03853662914773892,-0.15944597935464078,-0.08300493000858733,0.28182644674495094,-0.033492876063951904,0.10757193317796876,-0.10957921710819996,-0.30663870512074176,0.11237034968799442,-0.14524991378044305,0.26134915940576775,-0.10251891760705245,-0.1282171429883876,0.1586070626245064,-0.17405949758435213,-0.05813549521989447,-0.02974420906839903,-0.0255876906535768,0.03117852071154914,0.11968870089536932,-0.0769320554744386,0.08212723487432605,0.2335839002812925,0.22409858674452288,-0.20512317952714199,-0.07107431835072209,0.08872143522922116,0.24736177091653033,0.04078769867508214,0.06940239308333496,-0.04379622094108182,0.04920507195571133,0.14918415953437414,-0.12142287157032783,-2.2392816422467376E-5,-0.01273398884628918,0.1685449226359319,-0.12670552923365763,0.1260070239280351,-0.13306089197301585,0.036780451926656355,0.11219654191680561,-0.0868176360815891,0.2665435126276865,-0.06433617365977984,-0.014997997065353985,0.04535586890856269,-0.07788494276378,0.01927284695459619,0.04882607988723109,-0.16839293984837322,0.2649110756840214,0.10161708951354598,0.057618730094164405,0.5056188230851122,-3.1444925261325807E-4,0.18275148159096144,-0.0017763204585330428,-0.012994895825945818,-0.0021963936529783012,-0.020350544179204455,0.18663375786553438,0.11598911596006012,0.18550720954257383,-0.08938914586373592,-0.11191725996782331,0.17574155088060195,0.1394683890500774,-0.16849538140646647,-0.21057585588954275,-0.2530797083301213,-0.11497150831452885,0.01936732579467204,-0.12880480919231374,0.023923573403912712,-0.22776948246970186,0.010203230572978228,0.03598155762509059,-0.06118391446614435,-0.28431008181272777,0.164757398497823,-0.12279983532631444,2.1909155379606915E-4,0.03116626974959047,0.12442880171015691,0.19796405971254494,-0.11806682144006952,-0.20247921743267128,0.006707279001442856,-0.2804247027823994,0.034446030026538565,0.1932522761727471,-0.07695357096066147,0.2649900119132935,0.044054319265379505,0.24537455427813626,-0.03156364377314613,-0.019682102570590665,-0.05237048487114748,-0.2617840104722582,0.020181817141520675,-0.16618076816188232,0.16682101126917814,0.15595788785796305,0.15788687615842978,-0.1587173052283264,-0.011004964286196461,0.11048541474637652,-0.0599849008459179,0.02726234568348307,0.12877918972892388,-0.05441021032371918,0.039141029364874846,0.027279439731671573,-0.03688772958487041,-0.22907282684879782,-0.15579601920654262,0.02617482061585097,0.04175065778551623,-0.18498746048667858,-0.25244933955126037,0.030109872875955457,-0.020268884262851117,-0.024795107692837437,0.21075813727749093,-0.1491476095334745,0.21067877378083733,-0.07088417565983818,0.1305292505798361,-0.13076214139250109,-0.2344530367207387,-0.12198005531741174,-0.011230308583551007,0.22144580262800595,0.09815892869493315,0.03505118621561287,0.028649385255655524,0.06438627704068534,-0.039924707974160256,0.05377856967905803,-0.05454684578518609,-0.012504577168964026,-0.31776514253484667,0.049936571245631,-0.11157332414905863,-0.04201784375841555,-0.06267256813552446,0.1325681115135593,0.12101004866879528,-0.10350586654541782,-0.053882216606823796,0.18873330286696977,0.034630010769632236,0.08463797047215316,0.15375909752005068,-0.311743875074737,-0.02741265826803474,0.08978733019363982,0.14679078837098825,-0.06518895019847903,-0.16353015902198464,0.10694109036447852,0.17490059202340813,-0.05965042163794454,-0.10115838968731605,0.08706228181804132,-0.2901948379377357,0.07706112425649891,0.014130435811352982,-0.1519449705020124,-0.03927379237594408,-0.17000474609645622,-0.06441008107586921,0.19488634597347065,-0.23856732656663926,-0.15519074685930256,0.060951916951871965,-0.007016699444064578,0.0860295758005778,0.007136575445986892,0.07705192678697248,-0.2442974011309067,0.0341893071109407,-0.0012262481903198008,0.1462450278523315,-0.2565938483754824,0.09461034897802989,-0.1387777925578937,-0.1606770509591744,0.08874594342997977,-0.14761524344859644,-0.26720916343623635,-0.1565623506188647,-0.1699254510572974,0.0726404805912404,0.06111222726193155,-0.020146645413450237,-0.14099490371875678,-0.053174639681179414,0.027231389227213576,-0.11705098922890489,0.1657063825380689,0.0190609656570615,0.03498895969317406,0.15649918687757314,-0.05439783269719367,0.11155857338490094,-0.2539265019837389,0.09363800506369024,-0.09717467729201523,-0.08667425947771346,0.06021973437395012,-5.841006601112753E-4,0.04206985803899865,0.046259656205568156,0.1129224340288788,-0.31490691880692356,-0.10246615751528054,0.14009096625680684,-0.06881810796280785,-0.010544241313406033,0.11988517506873109,0.045641954476851154,-0.03610677266890816,-0.12489698563985563,0.0870474608048255,-0.2922526378859178,-0.16545417771181548,0.20160917865340444,-0.02601578706548402,0.037215858697476456,0.060368505087012636,-0.0424636376723031,-0.027844173558233543,0.3121276485925799,-0.27134771241371336,-0.2383189841961677,0.17827108118515422,0.03824116525050947,0.22939784738438007,0.032575769315102754,0.014234421010157483,-0.09042249121458656,0.005589690563333249,-0.0667483401074076,-0.3242449043904916,0.00851686378833616,0.3582604702046863,0.06756974555016448,0.050435475192064594,-0.07351877460863498,-0.11912181707383862,0.09928115055570252,0.016422897967386163,-0.13232850779156724,0.18375570421308618,-0.033142602707287416,0.07848253520732662,-0.0018341949735267704,0.08820482830538623,0.08776852433245086,0.1387363735948286,-0.06641478533771661,-0.23218199761579061,0.34251904609372197,0.1100700140161841,-0.08071360496460311,0.04013936911446825,0.035491352681638055,-0.04597799151486903,-0.004027860605411604,0.05773009392715078,0.22473392776414425,0.0970918077935723,-0.0643265823662098,-0.17451436679047294,0.054832532976986315,-0.08344362608819278,0.0636875330255755,-0.14121577436070298,-0.2684647672816336,0.11053817234005692,-0.0720845853818819,-0.11774358958531161,0.09105089725978743,-0.05422083848253445,0.1816791104955895,-0.040743443168982374,-0.19015413612899046,0.01411571784229615,0.12534988579931206,-0.15945666552532956,-0.09268969146969663,-0.10617249482245583,-0.009693164481050509,0.03371189289391374,-0.1855617677815493,-2.7206829355463924E-4,0.072030775607546,0.0016634329863771669,-0.08162405274634121,-0.02791519191354297,0.18127648329517232,-0.11590633559034469,-0.10956700093050298,0.07739941592115152,0.11767026295186067,-0.003247825930452919,0.02959770767732728,-0.10843347497132116,-0.05624546466667013,-0.15371559802984613,-0.09181582159136514,0.22461374325902117,-0.041938313322280596,-0.22537559798739792,0.0251648934995945,0.03425929289609897,0.03793250878226484,-0.19861797407591464,0.20711622027918253,0.06131055089137849,0.09278791651273188,-0.21118733833569703,0.08058107582653662,0.14220322397885887,0.003245634296051847,0.14558402597956022,-0.1533124323201891,-0.01206876618241104,0.23792001595592094,0.010012202282012161,0.028796532011563514,-0.06034547775246768,-0.03294680138983441,-0.0560597376355938,0.013160454492326501,0.16226959015477327,-0.30898855762309285,0.1556348934460683,0.03897123257461654,0.08046638268402351,-0.17979682079059156,-0.22438459238868624,-0.23599804060773355,0.07783131434922996,0.0301817225644192,0.02357815803162122,-0.032268304508297874,0.170111252522811,0.04398584329689026,-0.04114704799265567,-0.031232653033914648,-0.08684340114273592,0.03279953499274679,0.09637792973389114,0.0983827334347609,0.030606880810006842,-0.02308249065166379,0.05232317632182504,0.07146770402928926,-0.13674847353803887,-0.11751054438754593,-0.0824057377662637,-0.11186332696715162,-0.07954162218411831,0.06665410657697635,-0.1519352009520194,0.043670503892631274,-0.10627639276595957,-0.06536518269565317,-0.14838362928376903,-0.10256977505065024,-0.09304719778463166,0.04536500402289878,0.07290919030259957,-0.12698096530974912,-0.10612725855483607,0.09575254459060645,-0.03315622349545233,-0.03058951347418226,-0.1035773817595338,-0.24558062162201022,-0.14771704282221165,0.08806727830200245,-0.0638865244893676,0.32347217764430347,-0.02542423923665993,0.005930529182355926,0.029039823770665052,0.05483448549464498,-0.012258799543162446,-0.039779330413477254,-0.19513206747335676,0.13323265049552369,-0.021950827026818942,-0.11396076702143088,-0.1629516058247873,0.06921980889602697,-0.06654527693589873,-0.2659207861256494,0.01923026452197618,0.01746210003647697,-0.15170958273474167,-0.1304853479527397,0.10655386126221789,0.2864433749022899,0.18546423810623303,-0.09504998496977013,0.04130729794720178,0.1256406994674384,-0.06051916762691523,0.2389924005877706,0.15573948138615812,-0.03982591947817582,-0.04657083548395313,-0.040812295045288494,-0.045934024404035156,-0.017817455428673412,0.20934276396389784,-0.031227236054205854,0.04615167300053802,0.10084484530275706,-0.06800029745368086,0.11087628208912248,0.010658747996725282,-0.033488977759749794,-0.12027175285367253,0.20637923711764383,0.1682568328862848,0.007967543681568934,-0.22539572799774454,0.19783549686211188,0.12761182289252826,-0.1519954660412024,0.22704361588356292,0.03976071447946694,0.010525266288576418,0.2420198700795071,0.21482842159971138,-0.027981911874782384,-0.08170872294226061,0.06797791082147686,-0.14656518876239755,-0.16202073236087156,0.037407918408858265,-0.09856948520222275,0.09345280663342859,0.06358024656048666,-0.026551983660374923,-0.09940293296828084,0.10879478024963163,-0.29039016532479334,-0.0738113727085068,0.08532685551350294,0.0834014025085299,0.09584673624932805,0.14402829743999668,0.0459312931217184,-0.16855550383606221,0.0861866044024584,0.19483780712408016,0.015443180713611504,-0.12388911717386808,0.2250443780774134,0.21345488609596686,0.0190580090019683,-0.16335828508564543,0.09514592123715726,-0.2352927156205471,-0.2163247082973152,0.02807784154977533,0.04482018740529092,-0.2261032816289621,0.08143787153189486,-0.061812823639506576,-0.20885386522713478,0.1910572938340458,-0.016448774434858926,0.17826531904570686,0.10895946944822121,-0.06085069840193835,0.12363676102672727,0.14520939293988366,-0.04721786467853484,-0.0864878754614331,-0.0443474081863906,0.14497759533781895,-0.1636689296305116,0.1565785446794159,0.08325912648461209,0.055188652835850505,0.13367277343938877,0.16944801712939217,-0.014144130287895962,-0.01961290145444572,-0.056317834676079315,0.15140655220096733,0.04702325880466063,0.04292884952183851,0.1900024716617601,-0.05821886947892008,-0.058474818595316336,-0.34051720938409913,0.12268184056909526,-0.08713143183358252,0.05685720169099451,0.008881396975156441,-0.11639921972041556,0.10059813319577504,-0.1307632455589598,-0.17865601657344943,-0.06128386044014711,-0.10775517442886429,0.1201800023924573,0.2204981232384533,-0.0472329105113225,0.4034653548788061,0.016609846818409935,0.1480833211110273,-0.008195106517600692,-0.043418864755684794,0.18594553387490986,0.0841799007866566,-0.07597382006479059,-0.08896429644237358,-0.11105841508462522,-0.05521675018801009,0.14263101109504728,0.14172549906244036,0.18822197105548,0.25923421584091955,-0.20870856797892223,-0.0324105879532941,0.04460862362536311,0.14960213413081735,0.20139774914158082,0.27685558406801736,0.09616705179966445,-0.06308772489358727,-0.043381462514517466,-0.1717584575832305,-0.1636616379541591,-0.2619858601236751,-0.020136156070320155,0.13543846935379023,-0.0614416509490221,0.14091161148416376,-0.1296714754079833,0.3071922959080231,-0.2517994156761372,-0.07657406272720468,-0.08057407289516405,-0.2542847093069689,-0.2334028351723223,-0.1026988741197048,0.08935897438639628,-0.14453363622606652,0.09532334025443223,-0.08103426319661677,0.043245624741291344,-0.12060703076749935,0.07245974334910299,0.006056237405391407,0.024916886297981833,-0.2620615463267688,-0.06343505441672992,-0.15176675028098874,-0.0548895379403902,-0.16963437468960763,0.05557393274312794,0.06931152889018222,0.04771002591257023,-0.01935603278878162,0.208932989309762,0.1795071770222343,-9.437944854280289E-4,-0.04864413117993838,0.20718587467551597,-0.3590284127820764,-0.1280043814140024,0.08023809322887576,0.056617534711841166,0.025258544093106572,-0.11461148119551456,0.04886897526600542,-0.01408231725676945,-0.01514239379168636,-0.10209205311352143,-0.08489595907867731,0.03304372033027381,-0.3093099852465615,-0.13068409274873982,0.16866327626434163,0.012144690017032111,-0.141005620102149,0.21521531146682538,0.037099732655703654,-0.02266020042673933,-0.04156368145822675,-0.0701687198769605,0.04778570933850901,0.19038304607863932,-0.19396657911177417,0.08497631973304198,0.0409659410511658,0.26759488118725816,-0.14907219662029458,0.03363628623171788,-0.1202988194755461,0.14700860370735214,0.21121038336738973,-0.08289771185763595,0.1442814543121722,-0.12653720383083655,-0.044913467187183365,0.08057396657405569,-0.0927624392713442,-0.16882295320217366,-0.1486257457250897,-0.24843364775778465,0.2664232271254168,0.09253302755557868,0.12689849965202793,0.04056490085124054,0.07935338262410241,0.0369643817208507,0.15622192745934438,-0.16384120034759508,0.07514574419364095,0.13052981074960712,0.14617952443195215,-0.03113901870601321,0.14457385406892964,0.22851522363636612,-0.09038833640745736,-0.08703356658327532,-0.12396408714842622,0.0890482847755889,0.0028478410552637,0.18747023592961742,0.008335098544864916,-0.15543613511156718,-0.09210999565162048,-0.058804928593059635,0.02151521305302584,-0.3242973101567497,0.3150257557865954,0.14652663191712456,0.17704785486334362,0.17007620710083504,0.015196187385183801,0.14184657549500912,-0.007963526743306915,0.14405684238860772,0.09273589236162572,-0.04888783136466479,0.1737235181304131,-0.026468895585225027,-0.011724772674088335,0.011363787275520565,0.11130556237548639,0.12153130601025275,-0.09034104568612733,0.00972996644431312,0.11503221748765213,-0.223101906218663,-0.07522132649549951,-0.013257962241367771,-0.11950965880680275,0.11181217427297696,0.011738016543528573,0.16175099781536506,0.2075907802381627,0.16307981283116568,-0.10300221957851363,0.2157016037392463,0.09894418671288996,0.055138436810947766,0.1717262613749977,0.0047483669142653895,0.10184277369332762,0.029266072267561505,0.012518476556534708,0.03515180647220362,0.10568249543204544,0.09960404553399993,-0.03758222909117315,0.02019202709685608]\n",
              "},\n",
              "\"mapping\":{\n",
              "\"x\":\"predictions\",\n",
              "\"y\":\"residuals\"\n",
              "},\n",
              "\"data_meta\":{\n",
              "\"series_annotations\":[{\n",
              "\"type\":\"float\",\n",
              "\"column\":\"predictions\"\n",
              "},{\n",
              "\"type\":\"float\",\n",
              "\"column\":\"residuals\"\n",
              "}]\n",
              "},\n",
              "\"theme\":{\n",
              "\"name\":\"minimal\",\n",
              "\"panel_grid_minor\":{\n",
              "\"blank\":true\n",
              "}\n",
              "},\n",
              "\"guides\":{\n",
              "\"x\":{\n",
              "\"title\":\"Fitted values\"\n",
              "},\n",
              "\"y\":{\n",
              "\"title\":\"Residuals\"\n",
              "}\n",
              "},\n",
              "\"kind\":\"plot\",\n",
              "\"scales\":[{\n",
              "\"aesthetic\":\"y\",\n",
              "\"limits\":[-7.0,7.0]\n",
              "}],\n",
              "\"layers\":[{\n",
              "\"geom\":\"hline\",\n",
              "\"mapping\":{\n",
              "},\n",
              "\"data_meta\":{\n",
              "},\n",
              "\"yintercept\":0.0,\n",
              "\"linetype\":\"dashed\",\n",
              "\"data\":{\n",
              "}\n",
              "},{\n",
              "\"geom\":\"point\",\n",
              "\"mapping\":{\n",
              "},\n",
              "\"data_meta\":{\n",
              "},\n",
              "\"data\":{\n",
              "}\n",
              "}],\n",
              "\"metainfo_list\":[],\n",
              "\"spec_id\":\"4\"\n",
              "};\n",
              "               window.letsPlotCall(function() {\n",
              "       \n",
              "               var toolbar = null;\n",
              "               var plotContainer = containerDiv;               \n",
              "               \n",
              "                   var options = {\n",
              "                       sizing: {\n",
              "                           width_mode: \"min\",\n",
              "                           height_mode: \"scaled\",\n",
              "                           width: width\n",
              "                       }\n",
              "                   };\n",
              "                   var fig = LetsPlot.buildPlotFromProcessedSpecs(plotSpec, -1, -1, plotContainer, options);\n",
              "                   if (toolbar) {\n",
              "                     toolbar.bind(fig);\n",
              "                   }\n",
              "               });\n",
              "               \n",
              "               break;\n",
              "           }\n",
              "       }\n",
              "   });\n",
              "   \n",
              "   observer.observe(containerDiv);\n",
              "   \n",
              "   // ----------\n",
              "   })();\n",
              "   \n",
              "   </script>"
            ],
            "text/plain": [
              "<lets_plot.plot.core.PlotSpec at 0x1c4318d8380>"
            ]
          },
          "execution_count": 110,
          "metadata": {},
          "output_type": "execute_result"
        }
      ],
      "source": [
        "# Build a linear model\n",
        "sim_fit = linear_model.fit(v1.reshape(-1,1),v2.reshape(-1,1)) \n",
        "sim_preds = linear_model.predict(v1.reshape(-1,1))\n",
        "\n",
        "# Combine predictions / residuals to a data frame\n",
        "sim_residuals_toplot = pd.DataFrame({\"predictions\": sim_preds.reshape(-1),\n",
        "                                     \"residuals\": v2 - sim_preds.reshape(-1)})\n",
        "\n",
        "# Plot the results\n",
        "(\n",
        "  ggplot(sim_residuals_toplot, aes(\"predictions\", \"residuals\")) +\n",
        "  geom_hline(yintercept = 0, linetype = \"dashed\") +\n",
        "  geom_point() +\n",
        "  theme_minimal() +\n",
        "  theme(panel_grid_minor = element_blank()) +\n",
        "  scale_y_continuous(limits=[-7,7]) +\n",
        "  labs(x = \"Fitted values\", y = \"Residuals\")\n",
        ")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Now let's run this code for our model."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 21,
      "metadata": {},
      "outputs": [],
      "source": [
        "residuals_toplot = pd.DataFrame({\"predictions\": linear_preds,\n",
        "                                 \"residuals\": np.array(y_test) - linear_preds})\n",
        "\n",
        "# Plot the output of residuals_toplot (x-axis = predictions, y-axis = residuals)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "🎯 **ACTION POINTS** why does the graph of the simulated data illustrate a more well-fitting model when compared to our actual data?\n",
        "\n",
        "## Introduction to using list comprehensions to aid feature selection (30 minutes)\n",
        "\n",
        "👨🏻‍🏫 **TEACHING MOMENT:** Your tutor will take you through the code, so sit back, relax and enjoy!\n",
        "\n",
        "Remember our univariate model earlier? We are going to do the same for all features so see which ones show the best improvements in predictive power.\n",
        "\n",
        "We could build 15 different model objects, but this would be very inefficient. Instead, we are going to build a function that calculates the $R2$ for a given feature and apply it to all features using a list comprehension.\n",
        "\n",
        "### Create a list of feature names"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 112,
      "metadata": {},
      "outputs": [],
      "source": [
        "feature_names = wvs.drop([\"iso3c\",\"satisfaction\"],axis=1).columns"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Define a function to get r-squared values for each feature\n",
        "\n",
        "This gets a little tricky! Remember that we have transformed all our categorical features into one-hot encoded dummy variables. We need to therefore make sure that all dummies in one feature are included. The trick we've opted for is to use our original feature names to loop over.  In the `get_r2` we use a logical condition `X_train.columns.str.contains(feature)` which simply asks \"does a given column name contain the string value of a given feature?\" "
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 126,
      "metadata": {},
      "outputs": [],
      "source": [
        "def get_r2(feature):\n",
        "  \"\"\"\n",
        "  \n",
        "  Runs a linear model using a subset of features in a training set and calculates \n",
        "  the r-squared for each model, using a test set. The output is a single value.\n",
        "\n",
        "  \"\"\"\n",
        "  \n",
        "  # Create subsetted data\n",
        "  train = X_train[X_train.columns[X_train.columns.str.contains(feature)]]\n",
        "  test = X_test[X_test.columns[X_test.columns.str.contains(feature)]]\n",
        "  \n",
        "  # Instantiate and run a linear model on subsetted data\n",
        "  linear_model = LinearRegression()\n",
        "  linear_model.fit(train,y_train)\n",
        "  \n",
        "  # Create the predictions\n",
        "  preds = linear_model.predict(test)\n",
        "  \n",
        "  # Calculate r-squared\n",
        "  r2 = r2_score(y_test,preds)\n",
        "  return r2"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "We can check to see the function works by trying it on a feature or two.  Let's try this for financial situation."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 141,
      "metadata": {},
      "outputs": [
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "0.31\n"
          ]
        }
      ],
      "source": [
        "financial_situ_r2 = get_r2(\"financial_situ\")\n",
        "print(np.round(financial_situ_r2, 2))"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Use a list comprehension to loop `get_r2` over all features"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 140,
      "metadata": {},
      "outputs": [],
      "source": [
        "r2s = [get_r2(feat) for feat in feature_names]"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Create a data frame showing the r-squared for each feature"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 128,
      "metadata": {},
      "outputs": [],
      "source": [
        "linear_output = pd.DataFrame({\"feature\":feature_names,\"r2\":r2s})\n",
        "linear_output = linear_output.sort_values(by = \"r2\")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Plot the results!"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 125,
      "metadata": {},
      "outputs": [
        {
          "data": {
            "text/html": [
              "   <div id=\"bOm285\"></div>\n",
              "   <script type=\"text/javascript\" data-lets-plot-script=\"plot\">\n",
              "   \n",
              "   (function() {\n",
              "   // ----------\n",
              "   \n",
              "   var containerDiv = document.getElementById(\"bOm285\");\n",
              "   var observer = new ResizeObserver(function(entries) {\n",
              "       for (let entry of entries) {\n",
              "           var width = containerDiv.clientWidth\n",
              "           if (entry.contentBoxSize && width > 0) {\n",
              "           \n",
              "               // Render plot\n",
              "               if (observer) {\n",
              "                   observer.disconnect();\n",
              "                   observer = null;\n",
              "               }\n",
              "\n",
              "               var plotSpec={\n",
              "\"data\":{\n",
              "\"feature\":[\"rural\",\"male\",\"age\",\"post_second_edu\",\"relig_import\",\"married\",\"social_trust\",\"no_safety\",\"employment\",\"no_shelter\",\"no_medical\",\"no_food\",\"no_cash\",\"better_living\",\"financial_situ\"],\n",
              "\"r2\":[-8.85805217951674E-5,1.757759758702626E-5,1.2480110667700828E-4,0.001190160091574799,0.0031892982731426267,0.004203670116551983,0.004396858938114145,0.009379353364206922,0.010009704537274788,0.011206008970388814,0.022533902456326138,0.02461828835740687,0.027994425908696274,0.048674642753866015,0.3147842601900087]\n",
              "},\n",
              "\"mapping\":{\n",
              "\"x\":\"r2\",\n",
              "\"y\":\"feature\"\n",
              "},\n",
              "\"data_meta\":{\n",
              "\"series_annotations\":[{\n",
              "\"type\":\"str\",\n",
              "\"column\":\"feature\"\n",
              "},{\n",
              "\"type\":\"float\",\n",
              "\"column\":\"r2\"\n",
              "}]\n",
              "},\n",
              "\"theme\":{\n",
              "\"name\":\"minimal\",\n",
              "\"panel_grid_minor\":{\n",
              "\"blank\":true\n",
              "},\n",
              "\"panel_grid_major_y\":{\n",
              "\"blank\":true\n",
              "}\n",
              "},\n",
              "\"guides\":{\n",
              "\"x\":{\n",
              "\"title\":\"R-squared value\"\n",
              "},\n",
              "\"y\":{\n",
              "\"title\":\"\"\n",
              "}\n",
              "},\n",
              "\"kind\":\"plot\",\n",
              "\"scales\":[],\n",
              "\"layers\":[{\n",
              "\"geom\":\"bar\",\n",
              "\"stat\":\"identity\",\n",
              "\"mapping\":{\n",
              "},\n",
              "\"data_meta\":{\n",
              "},\n",
              "\"orientation\":\"y\",\n",
              "\"data\":{\n",
              "}\n",
              "}],\n",
              "\"metainfo_list\":[],\n",
              "\"spec_id\":\"8\"\n",
              "};\n",
              "               window.letsPlotCall(function() {\n",
              "       \n",
              "               var toolbar = null;\n",
              "               var plotContainer = containerDiv;               \n",
              "               \n",
              "                   var options = {\n",
              "                       sizing: {\n",
              "                           width_mode: \"min\",\n",
              "                           height_mode: \"scaled\",\n",
              "                           width: width\n",
              "                       }\n",
              "                   };\n",
              "                   var fig = LetsPlot.buildPlotFromProcessedSpecs(plotSpec, -1, -1, plotContainer, options);\n",
              "                   if (toolbar) {\n",
              "                     toolbar.bind(fig);\n",
              "                   }\n",
              "               });\n",
              "               \n",
              "               break;\n",
              "           }\n",
              "       }\n",
              "   });\n",
              "   \n",
              "   observer.observe(containerDiv);\n",
              "   \n",
              "   // ----------\n",
              "   })();\n",
              "   \n",
              "   </script>"
            ],
            "text/plain": [
              "<lets_plot.plot.core.PlotSpec at 0x1c4325a91f0>"
            ]
          },
          "execution_count": 125,
          "metadata": {},
          "output_type": "execute_result"
        }
      ],
      "source": [
        "(\n",
        "  ggplot(linear_output, aes(\"r2\", \"feature\")) +\n",
        "  geom_bar(stat = \"identity\") +\n",
        "  theme_minimal() +\n",
        "  theme(panel_grid_minor = element_blank(),\n",
        "        panel_grid_major_y = element_blank()) +\n",
        "  labs(x = \"R-squared value\", y = \"\")\n",
        ")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Using penalised linear regression to perform feature selection (20 minutes)\n",
        "\n",
        "We are now going to experiment with a lasso regression which, in this case, is a linear regression that uses a so-called **hyperparameter** - a \"dial\" built into a given model that can be experimented with to improve model performance. The hyperparameter in this case is a regularisation penalty which takes the value of a non-negative number. This penalty can shrink the magnitude of coefficients down to zero and the larger the penalty, the more shrinkage occurs.\n",
        "\n",
        "### Step 1: Create a lasso model\n",
        "\n",
        "Run the following code. This builds a lasso model with the penalty parameter set to `0.01`."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 33,
      "metadata": {},
      "outputs": [
        {
          "data": {
            "text/html": [
              "<style>#sk-container-id-2 {\n",
              "  /* Definition of color scheme common for light and dark mode */\n",
              "  --sklearn-color-text: #000;\n",
              "  --sklearn-color-text-muted: #666;\n",
              "  --sklearn-color-line: gray;\n",
              "  /* Definition of color scheme for unfitted estimators */\n",
              "  --sklearn-color-unfitted-level-0: #fff5e6;\n",
              "  --sklearn-color-unfitted-level-1: #f6e4d2;\n",
              "  --sklearn-color-unfitted-level-2: #ffe0b3;\n",
              "  --sklearn-color-unfitted-level-3: chocolate;\n",
              "  /* Definition of color scheme for fitted estimators */\n",
              "  --sklearn-color-fitted-level-0: #f0f8ff;\n",
              "  --sklearn-color-fitted-level-1: #d4ebff;\n",
              "  --sklearn-color-fitted-level-2: #b3dbfd;\n",
              "  --sklearn-color-fitted-level-3: cornflowerblue;\n",
              "\n",
              "  /* Specific color for light theme */\n",
              "  --sklearn-color-text-on-default-background: var(--sg-text-color, var(--theme-code-foreground, var(--jp-content-font-color1, black)));\n",
              "  --sklearn-color-background: var(--sg-background-color, var(--theme-background, var(--jp-layout-color0, white)));\n",
              "  --sklearn-color-border-box: var(--sg-text-color, var(--theme-code-foreground, var(--jp-content-font-color1, black)));\n",
              "  --sklearn-color-icon: #696969;\n",
              "\n",
              "  @media (prefers-color-scheme: dark) {\n",
              "    /* Redefinition of color scheme for dark theme */\n",
              "    --sklearn-color-text-on-default-background: var(--sg-text-color, var(--theme-code-foreground, var(--jp-content-font-color1, white)));\n",
              "    --sklearn-color-background: var(--sg-background-color, var(--theme-background, var(--jp-layout-color0, #111)));\n",
              "    --sklearn-color-border-box: var(--sg-text-color, var(--theme-code-foreground, var(--jp-content-font-color1, white)));\n",
              "    --sklearn-color-icon: #878787;\n",
              "  }\n",
              "}\n",
              "\n",
              "#sk-container-id-2 {\n",
              "  color: var(--sklearn-color-text);\n",
              "}\n",
              "\n",
              "#sk-container-id-2 pre {\n",
              "  padding: 0;\n",
              "}\n",
              "\n",
              "#sk-container-id-2 input.sk-hidden--visually {\n",
              "  border: 0;\n",
              "  clip: rect(1px 1px 1px 1px);\n",
              "  clip: rect(1px, 1px, 1px, 1px);\n",
              "  height: 1px;\n",
              "  margin: -1px;\n",
              "  overflow: hidden;\n",
              "  padding: 0;\n",
              "  position: absolute;\n",
              "  width: 1px;\n",
              "}\n",
              "\n",
              "#sk-container-id-2 div.sk-dashed-wrapped {\n",
              "  border: 1px dashed var(--sklearn-color-line);\n",
              "  margin: 0 0.4em 0.5em 0.4em;\n",
              "  box-sizing: border-box;\n",
              "  padding-bottom: 0.4em;\n",
              "  background-color: var(--sklearn-color-background);\n",
              "}\n",
              "\n",
              "#sk-container-id-2 div.sk-container {\n",
              "  /* jupyter's `normalize.less` sets `[hidden] { display: none; }`\n",
              "     but bootstrap.min.css set `[hidden] { display: none !important; }`\n",
              "     so we also need the `!important` here to be able to override the\n",
              "     default hidden behavior on the sphinx rendered scikit-learn.org.\n",
              "     See: https://github.com/scikit-learn/scikit-learn/issues/21755 */\n",
              "  display: inline-block !important;\n",
              "  position: relative;\n",
              "}\n",
              "\n",
              "#sk-container-id-2 div.sk-text-repr-fallback {\n",
              "  display: none;\n",
              "}\n",
              "\n",
              "div.sk-parallel-item,\n",
              "div.sk-serial,\n",
              "div.sk-item {\n",
              "  /* draw centered vertical line to link estimators */\n",
              "  background-image: linear-gradient(var(--sklearn-color-text-on-default-background), var(--sklearn-color-text-on-default-background));\n",
              "  background-size: 2px 100%;\n",
              "  background-repeat: no-repeat;\n",
              "  background-position: center center;\n",
              "}\n",
              "\n",
              "/* Parallel-specific style estimator block */\n",
              "\n",
              "#sk-container-id-2 div.sk-parallel-item::after {\n",
              "  content: \"\";\n",
              "  width: 100%;\n",
              "  border-bottom: 2px solid var(--sklearn-color-text-on-default-background);\n",
              "  flex-grow: 1;\n",
              "}\n",
              "\n",
              "#sk-container-id-2 div.sk-parallel {\n",
              "  display: flex;\n",
              "  align-items: stretch;\n",
              "  justify-content: center;\n",
              "  background-color: var(--sklearn-color-background);\n",
              "  position: relative;\n",
              "}\n",
              "\n",
              "#sk-container-id-2 div.sk-parallel-item {\n",
              "  display: flex;\n",
              "  flex-direction: column;\n",
              "}\n",
              "\n",
              "#sk-container-id-2 div.sk-parallel-item:first-child::after {\n",
              "  align-self: flex-end;\n",
              "  width: 50%;\n",
              "}\n",
              "\n",
              "#sk-container-id-2 div.sk-parallel-item:last-child::after {\n",
              "  align-self: flex-start;\n",
              "  width: 50%;\n",
              "}\n",
              "\n",
              "#sk-container-id-2 div.sk-parallel-item:only-child::after {\n",
              "  width: 0;\n",
              "}\n",
              "\n",
              "/* Serial-specific style estimator block */\n",
              "\n",
              "#sk-container-id-2 div.sk-serial {\n",
              "  display: flex;\n",
              "  flex-direction: column;\n",
              "  align-items: center;\n",
              "  background-color: var(--sklearn-color-background);\n",
              "  padding-right: 1em;\n",
              "  padding-left: 1em;\n",
              "}\n",
              "\n",
              "\n",
              "/* Toggleable style: style used for estimator/Pipeline/ColumnTransformer box that is\n",
              "clickable and can be expanded/collapsed.\n",
              "- Pipeline and ColumnTransformer use this feature and define the default style\n",
              "- Estimators will overwrite some part of the style using the `sk-estimator` class\n",
              "*/\n",
              "\n",
              "/* Pipeline and ColumnTransformer style (default) */\n",
              "\n",
              "#sk-container-id-2 div.sk-toggleable {\n",
              "  /* Default theme specific background. It is overwritten whether we have a\n",
              "  specific estimator or a Pipeline/ColumnTransformer */\n",
              "  background-color: var(--sklearn-color-background);\n",
              "}\n",
              "\n",
              "/* Toggleable label */\n",
              "#sk-container-id-2 label.sk-toggleable__label {\n",
              "  cursor: pointer;\n",
              "  display: flex;\n",
              "  width: 100%;\n",
              "  margin-bottom: 0;\n",
              "  padding: 0.5em;\n",
              "  box-sizing: border-box;\n",
              "  text-align: center;\n",
              "  align-items: start;\n",
              "  justify-content: space-between;\n",
              "  gap: 0.5em;\n",
              "}\n",
              "\n",
              "#sk-container-id-2 label.sk-toggleable__label .caption {\n",
              "  font-size: 0.6rem;\n",
              "  font-weight: lighter;\n",
              "  color: var(--sklearn-color-text-muted);\n",
              "}\n",
              "\n",
              "#sk-container-id-2 label.sk-toggleable__label-arrow:before {\n",
              "  /* Arrow on the left of the label */\n",
              "  content: \"▸\";\n",
              "  float: left;\n",
              "  margin-right: 0.25em;\n",
              "  color: var(--sklearn-color-icon);\n",
              "}\n",
              "\n",
              "#sk-container-id-2 label.sk-toggleable__label-arrow:hover:before {\n",
              "  color: var(--sklearn-color-text);\n",
              "}\n",
              "\n",
              "/* Toggleable content - dropdown */\n",
              "\n",
              "#sk-container-id-2 div.sk-toggleable__content {\n",
              "  max-height: 0;\n",
              "  max-width: 0;\n",
              "  overflow: hidden;\n",
              "  text-align: left;\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-0);\n",
              "}\n",
              "\n",
              "#sk-container-id-2 div.sk-toggleable__content.fitted {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-0);\n",
              "}\n",
              "\n",
              "#sk-container-id-2 div.sk-toggleable__content pre {\n",
              "  margin: 0.2em;\n",
              "  border-radius: 0.25em;\n",
              "  color: var(--sklearn-color-text);\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-0);\n",
              "}\n",
              "\n",
              "#sk-container-id-2 div.sk-toggleable__content.fitted pre {\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-0);\n",
              "}\n",
              "\n",
              "#sk-container-id-2 input.sk-toggleable__control:checked~div.sk-toggleable__content {\n",
              "  /* Expand drop-down */\n",
              "  max-height: 200px;\n",
              "  max-width: 100%;\n",
              "  overflow: auto;\n",
              "}\n",
              "\n",
              "#sk-container-id-2 input.sk-toggleable__control:checked~label.sk-toggleable__label-arrow:before {\n",
              "  content: \"▾\";\n",
              "}\n",
              "\n",
              "/* Pipeline/ColumnTransformer-specific style */\n",
              "\n",
              "#sk-container-id-2 div.sk-label input.sk-toggleable__control:checked~label.sk-toggleable__label {\n",
              "  color: var(--sklearn-color-text);\n",
              "  background-color: var(--sklearn-color-unfitted-level-2);\n",
              "}\n",
              "\n",
              "#sk-container-id-2 div.sk-label.fitted input.sk-toggleable__control:checked~label.sk-toggleable__label {\n",
              "  background-color: var(--sklearn-color-fitted-level-2);\n",
              "}\n",
              "\n",
              "/* Estimator-specific style */\n",
              "\n",
              "/* Colorize estimator box */\n",
              "#sk-container-id-2 div.sk-estimator input.sk-toggleable__control:checked~label.sk-toggleable__label {\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-2);\n",
              "}\n",
              "\n",
              "#sk-container-id-2 div.sk-estimator.fitted input.sk-toggleable__control:checked~label.sk-toggleable__label {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-2);\n",
              "}\n",
              "\n",
              "#sk-container-id-2 div.sk-label label.sk-toggleable__label,\n",
              "#sk-container-id-2 div.sk-label label {\n",
              "  /* The background is the default theme color */\n",
              "  color: var(--sklearn-color-text-on-default-background);\n",
              "}\n",
              "\n",
              "/* On hover, darken the color of the background */\n",
              "#sk-container-id-2 div.sk-label:hover label.sk-toggleable__label {\n",
              "  color: var(--sklearn-color-text);\n",
              "  background-color: var(--sklearn-color-unfitted-level-2);\n",
              "}\n",
              "\n",
              "/* Label box, darken color on hover, fitted */\n",
              "#sk-container-id-2 div.sk-label.fitted:hover label.sk-toggleable__label.fitted {\n",
              "  color: var(--sklearn-color-text);\n",
              "  background-color: var(--sklearn-color-fitted-level-2);\n",
              "}\n",
              "\n",
              "/* Estimator label */\n",
              "\n",
              "#sk-container-id-2 div.sk-label label {\n",
              "  font-family: monospace;\n",
              "  font-weight: bold;\n",
              "  display: inline-block;\n",
              "  line-height: 1.2em;\n",
              "}\n",
              "\n",
              "#sk-container-id-2 div.sk-label-container {\n",
              "  text-align: center;\n",
              "}\n",
              "\n",
              "/* Estimator-specific */\n",
              "#sk-container-id-2 div.sk-estimator {\n",
              "  font-family: monospace;\n",
              "  border: 1px dotted var(--sklearn-color-border-box);\n",
              "  border-radius: 0.25em;\n",
              "  box-sizing: border-box;\n",
              "  margin-bottom: 0.5em;\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-0);\n",
              "}\n",
              "\n",
              "#sk-container-id-2 div.sk-estimator.fitted {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-0);\n",
              "}\n",
              "\n",
              "/* on hover */\n",
              "#sk-container-id-2 div.sk-estimator:hover {\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-2);\n",
              "}\n",
              "\n",
              "#sk-container-id-2 div.sk-estimator.fitted:hover {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-2);\n",
              "}\n",
              "\n",
              "/* Specification for estimator info (e.g. \"i\" and \"?\") */\n",
              "\n",
              "/* Common style for \"i\" and \"?\" */\n",
              "\n",
              ".sk-estimator-doc-link,\n",
              "a:link.sk-estimator-doc-link,\n",
              "a:visited.sk-estimator-doc-link {\n",
              "  float: right;\n",
              "  font-size: smaller;\n",
              "  line-height: 1em;\n",
              "  font-family: monospace;\n",
              "  background-color: var(--sklearn-color-background);\n",
              "  border-radius: 1em;\n",
              "  height: 1em;\n",
              "  width: 1em;\n",
              "  text-decoration: none !important;\n",
              "  margin-left: 0.5em;\n",
              "  text-align: center;\n",
              "  /* unfitted */\n",
              "  border: var(--sklearn-color-unfitted-level-1) 1pt solid;\n",
              "  color: var(--sklearn-color-unfitted-level-1);\n",
              "}\n",
              "\n",
              ".sk-estimator-doc-link.fitted,\n",
              "a:link.sk-estimator-doc-link.fitted,\n",
              "a:visited.sk-estimator-doc-link.fitted {\n",
              "  /* fitted */\n",
              "  border: var(--sklearn-color-fitted-level-1) 1pt solid;\n",
              "  color: var(--sklearn-color-fitted-level-1);\n",
              "}\n",
              "\n",
              "/* On hover */\n",
              "div.sk-estimator:hover .sk-estimator-doc-link:hover,\n",
              ".sk-estimator-doc-link:hover,\n",
              "div.sk-label-container:hover .sk-estimator-doc-link:hover,\n",
              ".sk-estimator-doc-link:hover {\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-3);\n",
              "  color: var(--sklearn-color-background);\n",
              "  text-decoration: none;\n",
              "}\n",
              "\n",
              "div.sk-estimator.fitted:hover .sk-estimator-doc-link.fitted:hover,\n",
              ".sk-estimator-doc-link.fitted:hover,\n",
              "div.sk-label-container:hover .sk-estimator-doc-link.fitted:hover,\n",
              ".sk-estimator-doc-link.fitted:hover {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-3);\n",
              "  color: var(--sklearn-color-background);\n",
              "  text-decoration: none;\n",
              "}\n",
              "\n",
              "/* Span, style for the box shown on hovering the info icon */\n",
              ".sk-estimator-doc-link span {\n",
              "  display: none;\n",
              "  z-index: 9999;\n",
              "  position: relative;\n",
              "  font-weight: normal;\n",
              "  right: .2ex;\n",
              "  padding: .5ex;\n",
              "  margin: .5ex;\n",
              "  width: min-content;\n",
              "  min-width: 20ex;\n",
              "  max-width: 50ex;\n",
              "  color: var(--sklearn-color-text);\n",
              "  box-shadow: 2pt 2pt 4pt #999;\n",
              "  /* unfitted */\n",
              "  background: var(--sklearn-color-unfitted-level-0);\n",
              "  border: .5pt solid var(--sklearn-color-unfitted-level-3);\n",
              "}\n",
              "\n",
              ".sk-estimator-doc-link.fitted span {\n",
              "  /* fitted */\n",
              "  background: var(--sklearn-color-fitted-level-0);\n",
              "  border: var(--sklearn-color-fitted-level-3);\n",
              "}\n",
              "\n",
              ".sk-estimator-doc-link:hover span {\n",
              "  display: block;\n",
              "}\n",
              "\n",
              "/* \"?\"-specific style due to the `<a>` HTML tag */\n",
              "\n",
              "#sk-container-id-2 a.estimator_doc_link {\n",
              "  float: right;\n",
              "  font-size: 1rem;\n",
              "  line-height: 1em;\n",
              "  font-family: monospace;\n",
              "  background-color: var(--sklearn-color-background);\n",
              "  border-radius: 1rem;\n",
              "  height: 1rem;\n",
              "  width: 1rem;\n",
              "  text-decoration: none;\n",
              "  /* unfitted */\n",
              "  color: var(--sklearn-color-unfitted-level-1);\n",
              "  border: var(--sklearn-color-unfitted-level-1) 1pt solid;\n",
              "}\n",
              "\n",
              "#sk-container-id-2 a.estimator_doc_link.fitted {\n",
              "  /* fitted */\n",
              "  border: var(--sklearn-color-fitted-level-1) 1pt solid;\n",
              "  color: var(--sklearn-color-fitted-level-1);\n",
              "}\n",
              "\n",
              "/* On hover */\n",
              "#sk-container-id-2 a.estimator_doc_link:hover {\n",
              "  /* unfitted */\n",
              "  background-color: var(--sklearn-color-unfitted-level-3);\n",
              "  color: var(--sklearn-color-background);\n",
              "  text-decoration: none;\n",
              "}\n",
              "\n",
              "#sk-container-id-2 a.estimator_doc_link.fitted:hover {\n",
              "  /* fitted */\n",
              "  background-color: var(--sklearn-color-fitted-level-3);\n",
              "}\n",
              "</style><div id=\"sk-container-id-2\" class=\"sk-top-container\"><div class=\"sk-text-repr-fallback\"><pre>Lasso(alpha=0.01)</pre><b>In a Jupyter environment, please rerun this cell to show the HTML representation or trust the notebook. <br />On GitHub, the HTML representation is unable to render, please try loading this page with nbviewer.org.</b></div><div class=\"sk-container\" hidden><div class=\"sk-item\"><div class=\"sk-estimator fitted sk-toggleable\"><input class=\"sk-toggleable__control sk-hidden--visually\" id=\"sk-estimator-id-2\" type=\"checkbox\" checked><label for=\"sk-estimator-id-2\" class=\"sk-toggleable__label fitted sk-toggleable__label-arrow\"><div><div>Lasso</div></div><div><a class=\"sk-estimator-doc-link fitted\" rel=\"noreferrer\" target=\"_blank\" href=\"https://scikit-learn.org/1.6/modules/generated/sklearn.linear_model.Lasso.html\">?<span>Documentation for Lasso</span></a><span class=\"sk-estimator-doc-link fitted\">i<span>Fitted</span></span></div></label><div class=\"sk-toggleable__content fitted\"><pre>Lasso(alpha=0.01)</pre></div> </div></div></div></div>"
            ],
            "text/plain": [
              "Lasso(alpha=0.01)"
            ]
          },
          "execution_count": 33,
          "metadata": {},
          "output_type": "execute_result"
        }
      ],
      "source": [
        "lasso_model = Lasso(alpha = 0.01)\n",
        "lasso_model.fit(X_train, y_train)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Step 2: Extract lasso coefficients"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 34,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Create a data frame with feature columns and Lasso coefficients\n",
        "lasso_output = pd.DataFrame({\"feature\": X_train.columns, \"coefficient\": lasso_model.coef_})\n",
        "\n",
        "# Code a positive / negative vector\n",
        "lasso_output[\"positive\"] = np.where(lasso_output[\"coefficient\"] >= 0, True, False)\n",
        "\n",
        "# Take the absolute value of the coefficients\n",
        "lasso_output[\"coefficient\"] = np.abs(lasso_output[\"coefficient\"])\n",
        "\n",
        "# Remove the constant and sort the data frame by (absolute) coefficient magnitude\n",
        "lasso_output = lasso_output.query(\"feature != 'const'\").sort_values(\"coefficient\")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "🎯 **ACTION POINTS** What is the output? Which coefficients have been shrunk to zero? What is the most important feature?\n",
        "\n",
        "### Step 3: Create a bar plot"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 22,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Code here"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Step 4: Evaluate on the test set\n",
        "\n",
        "Although a different model is used, the code for evaluating the model on the test set is exactly the same as earlier."
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 36,
      "metadata": {},
      "outputs": [
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "0.33 1.81\n"
          ]
        }
      ],
      "source": [
        "# Apply model to test set\n",
        "\n",
        "# Calculate performance metrics\n",
        "\n",
        "# Print rounded PMs\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "🗣️ **CLASSROOM DISCUSSION:**\n",
        "\n",
        "What feature is the largest positive / negative predictor of life satisfaction?  Does this model represent an improvement on the linear model?\n",
        "\n",
        "### (Bonus) Step 5: Experiment with different penalties\n",
        "\n",
        "This is your chance to try out different penalties. Can you find a penalty that improves test set performance?"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 23,
      "metadata": {},
      "outputs": [],
      "source": [
        "# Instantiate a lasso model\n",
        "\n",
        "\n",
        "# Fit the model to the training data\n",
        "\n",
        "\n",
        "# Apply the model to the test set\n",
        "\n",
        "\n",
        "# Calculate performance metrics\n",
        "\n",
        "\n",
        "# Print the rounded metrics\n",
        "\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "👉 **NOTE:** In labs 4 and 5, we are going to use a method called k-fold cross validation to systematically test different combinations of hyperparameters for models such as the lasso."
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.12.3"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 4
}
